//$Id$


/**
 * @class
 * A <b>static</b> class containing API's for  Ajax related operations.
 * It has methods for 
 * <ul><li> Sending Ajax Requests
 *        <ul><li>{@link #sendRequest}</li></ul></li>
 *     <li> Predefined functions for handling response such as
 *        <ul><li>{@link #closeView}</li>
 *            <li>{@link #refreshView}</li></ul></li>
 *     <li> General Utilities such as
 *         <ul><li>{@link #setOnNextPageLoadScript}</li></ul></li>
 *
 *            
 *<br/>
 * @see WebViewAPI.sendResponse in JavaAPI.
 */
var AjaxAPI = new function(){}

/**
 * The name of the iframe to use in {@link AjaxAPI#sendNavigableRequest}.
 * @type String
 */
AjaxAPI.navigFrameName = null;  
AjaxAPI.navigFrameEl = null;  


  /**
   * Returns the native implementation for XMLHTTP based on browser.
   * Developers should avoid using this API. Rather sendRequest should be used.
   * @exception Error
   * @return returns XMLHttpRequest (or) ActiveXObject
   * @see #sendRequest
   */
AjaxAPI.getXMLHttpRequest = function()
{
  try
  {
    var http_request = null;
    if (window.XMLHttpRequest) 
    { // Mozilla, Safari,...
      http_request = new XMLHttpRequest();
    } 
    else if (window.ActiveXObject) 
    { // IE
      try 
      {
        http_request = new ActiveXObject("Msxml2.XMLHTTP");
      } 
      catch (e) 
      {
        http_request = new ActiveXObject("Microsoft.XMLHTTP");
      }
    }
  }
  catch(e)
  {
    StatusMsgAPI.showErrorMsg("Unable to create XML Http Object for Ajax Request. <br>Reason :" + LangUtils.getErrorMsg(e));
    throw e;
  }
  return http_request;
}

/**
 * Use this function to send an Ajax Request.
 *<br>  Following are taken care of.
 *<ul>
 *    <li>Showing loading/loaded operation status</li>
 *    <li>Showing Success/Error Message .  </li>
 *    <li>Showing Error Message incase of server error.</li>
 *    <li>Showing a mini-login screen incase of session time out.</li>
 *    <li>Invoking of scripts if present in response.(Will be done at the end after callback function
 *      is called).</li>
 *</ul>
 *<h4>Sample 1</h4>
 *<pre>  <code>AjaxAPI.sendRequest({URL:"MyDeleteActionAction.ma",PARAMETERS:"RESID=1"});</code></pre>
 *<h4>Using Predefined Callback Functions</h4>
 *<pre>  <code> AjaxAPI.sendRequest({URL:"MyDeleteAction.ma",PARAMETERS:"RESID=1",ONSUCCESSFUNC:"AjaxAPI.refreshView",SRCVIEW:"MyView"});</code>  </pre>
 *<h4>Using anonyamous/inner call back functions</h4>
 *<pre><code>
 *  function myTestFunc()
 *  {
 *    var <span style="color:green">localVar</span> = "HelloWorldDiv";
 *    var <span style="color:blue">myAnonFunction</span> = function(response,reqOptions)
 *    {
 *       alert("Wow. Succeeded."); 
 *       alert(response.responseText);
 *       document.getElementById(<span style="color:green">localVar</span>).innerHTML = <b>response.getOnlyHtml();</b>
 *
 *       //Not really needed in this case as local variables are accessable even though this function 
 *       //will be called asynchronously. Custom reqOptions are usefull for reusable functions such as
 *       //AjaxAPI.refreshView!!
 *       alert(reqOptions["MYCALLBACKPARAM"]);
 *    };
 *    AjaxAPI.sendRequest({URL:"MyDeleteAction.ma",PARAMETERS:"RESID=1",MYCALLBACKPARAM:"HAI",ONSUCCESSFUNC:<span style="color:blue">myAnonFunction</span>});
 *  }
 * </code></pre>
 *
 * @param {Object} requestOptions An JavaScript Object  containing the various options. This set of parameters are passed back
 * to the callback functions along with response.
 * <h3>Main Option details :</h3>
 * <ul><li><code>URL</code> <ul><li>The url to send </li><li><i>Mandatory</i></li></ul></li>
 * <li><code>METHOD</code>  <ul><li>  The protocol method to use. </li>
 *                             <li><i>Allowed</i> : GET/POST</li><li><i>Default :</i> POST</li>
 *                         </ul> </li>
 * <li><code>PARAMETERS</code> <ul><li>The parameters to send. For POST , it is sent in request body. 
 *                            For GET , the parameters are attached to the URL itself.</li></ul></li>
 * <li><code>STATUSFUNC </code> <ul><li>The function to invoke for showing 
 *         status and handling response messages.Generally this is not overridden.
 *         <li><i>Default :</i> {@link AjaxAPI#showRespMsg}</li><li><i>Data Type</i>: function name (or) reference to a function</ul></li>
 * <li><code>ONSUCCESSFUNC</code> : <ul><li>The function to invoke in case of success.See {@link AjaxAPI#refreshView}
 *  for details about the method signature</li><li><i>Data Type</i> : function name (or) reference to a function</ul></li></ul></li>
 * <li><code>ONFAILUREFUNC</code> : <ul><li>The function to invoke in case of failure.Similar to ONSUCCESSFUNC</li></ul></li>
 * </ul>
 */
AjaxAPI.sendRequest = function(requestOptions)
{
  var xmlHttpReq = AjaxAPI.getXMLHttpRequest();
  try
  {
    requestOptions = AjaxOptions.getAsAjaxOptions(requestOptions);

    var reqURL = requestOptions["URL"];
        
    var parameters = requestOptions["PARAMETERS"];
    var url = requestOptions.v("URL");
    if(url.indexOf(location.protocol) != 0 && url.indexOf("/reports/") != 0)
    {
      url = ZDBGlobal.REPORTS_CONTEXT_NAME + url;
    }
    var target = requestOptions["TARGET"];
          
    if (parameters && (requestOptions.vu("METHOD") == 'GET'))
    {
      url = appendParamsToUrl(url, parameters);
    } 

    //We have to append this because even though we will be able to get the 
    // info from request header, tomcat does not pass this info properly during session time relogin!!.
    // See WebViewAPI.isAjaxRequest!!.
    if(!requestOptions["NOSUBREQPARAM"])
    {
      url = appendParamsToUrl(url,"SUBREQUEST=XMLHTTP&_ZVER_="+ZDBGlobal._ZVER_);// No I18n
    }
    
//    url = AjaxUtils.addAdditionalParamsForAuthenticationSkip(url);

    xmlHttpReq.open(requestOptions.vu("METHOD"), url,
                    requestOptions.isTrue("ASYNCHRONOUS"));
    
    var requestHeaders = new Object();
    requestHeaders["X-Requested-With"] = 'XMLHttpRequest';

    if (requestOptions.vu("METHOD") == 'POST') 
    {
      requestHeaders['Content-type']= 'application/x-www-form-urlencoded;charset=UTF-8';
    }

    //if (xmlHttpReq.overrideMimeType)
    //{
    //  requestHeaders['Connection'] = 'close';
    //}
    AjaxUtils.updateObject(requestHeaders, requestOptions["REQUESTHEADERS"]);

    if((parameters != null) && (requestOptions.vu("METHOD") == 'POST'))
    {
      parameters += "&ZDB_CSRF_TOKEN=" + getCookie("CSRF_TOKEN");//NO I18N
    }

    var body = requestOptions["POSTBODY"] ? requestOptions["POSTBODY"] : parameters;
    body = (requestOptions.vu("METHOD") == 'POST')? body : null;
    if(body != null)
    {
      requestHeaders['Content-length',body.length];
    }
    else if(requestOptions.vu("METHOD") == 'POST')
    {
      body="___DUMMYDATA=aa&ZDB_CSRF_TOKEN=" + getCookie("CSRF_TOKEN");//NO I18N
      requestHeaders['Content-length',body.length];
    }

    for (var header in requestHeaders)
    {
      xmlHttpReq.setRequestHeader(header,requestHeaders[header]);
    }
    
    var resp = new AjaxResponse(xmlHttpReq);
    xmlHttpReq.onreadystatechange = function(){AjaxUtils.handleAjaxResponse(resp,requestOptions);};

    if(requestOptions["STATUSFUNC"])
    {
      requestOptions.fn("STATUSFUNC")(resp,requestOptions);
    }
    
    /**** Progress bar handling start *******/
    if(requestOptions.PROGRESSIND === "true")
  {
      var upload = xmlHttpReq.upload;
        upload.downloadStartTime = new Date().getTime();
        upload.currentStart = upload.downloadStartTime;
        upload.currentProgress = 0;
        upload.startData = 0;
        upload.addEventListener("progress", function(event){AjaxAPI.updateProgress(event,requestOptions);});
        upload.addEventListener("load", function(event){AjaxAPI.completeProgress(event,requestOptions);});
  }
    /**** Progress bar handling End *******/
   
   /**** Initialise Progress Bar Handled by MONITORFUNC *******/ 
   if(requestOptions["INITFUNC"]){ //No I18N
   	 requestOptions.fn("INITFUNC")(resp, requestOptions); //No I18N
   }
   /**** Initialisation of Progress Bar Ends  *******/
    
    xmlHttpReq.send(body);
    DOMUtils.enableElements(requestOptions["DISABLEELS"],false);
  }
  catch(e)
  {
    StatusMsgAPI.showErrorMsg("Unable to send request.<br>Reason : " + LangUtils.getErrorMsg(e));
    throw e;
  }
}


/**
 * Similar to {@link AjaxAPI#sendRequest}. But the Back/Next Button would be supported as
 * the request is sent via an iframe.
 * <br><b>Note:</b>
 *  <ul><li>Only GET is currently supported</li>
 *      <li>Ensure that <b>bare minimal</b> options are used as these options are also encoded
 *         into the URL as hash part. More options, more the size of the url!!
 *      <ul><li>See {@link AjaxOptions#encode}</li></ul></li></ul>
 *
 * @param {Map} requestOptions
 */
AjaxAPI.sendNavigableRequest = function(requestOptions)
{
  requestOptions = AjaxOptions.getAsAjaxOptions(requestOptions);
  if(requestOptions["METHOD"] != "GET"){throw new Error("Currently only Get is supported.");}
  var url = AjaxAPI.getAsNavigableAjaxURL(requestOptions);
  if(AjaxAPI.navigFrameName != null)
  {
    AjaxAPI.navigFrameEl.reqOptions = null;
    window.frames[AjaxAPI.navigFrameName].location.href = url;
  }
  else
  {
    throw new Error("AjaxAPI.navigFrameName has not been set");
  }
}

/**
 * Returns MCFrame object for the contentareaname 
 * detail present in requestOptions
 * @param {Map} requestOptions
 */
AjaxAPI.getMCFrame = function(requestOptions)
{
  if(requestOptions["CONTENTAREANAME"] != null)
  {
    return MCFrame.getMCFrame(requestOptions["CONTENTAREANAME"]);
  }
    else if(requestOptions["NAME"] != null 
            && requestOptions["TARGET"].indexOf("_mcframe_this") != -1)
    {
        var frm = DOMUtils.getForm(requestOptions["NAME"]);
        var mcframe = MCFrame.getMCFrame(frm);
        requestOptions["CONTENTAREANAME"] = mcframe.contentareaname;
        return mcframe;
    }
    else
    {
        var target = requestOptions["TARGET"];
        var contentAreaName = target.substring("_mcframe_".length);
        if(contentAreaName.indexOf("_") != -1)
        {
          contentAreaName = contentAreaName.substring(0, contentAreaName.indexOf("_"));
        }
        return MCFrame.getMCFrame(contentAreaName);
    } 
}

AjaxAPI.getAsNavigableAjaxURL = function(requestOptions)
{
  requestOptions = AjaxOptions.getAsAjaxOptions(requestOptions);
  var encodedOptions = requestOptions.encode(); 
  var url = requestOptions["URL"];
  if(url.indexOf("SUBREQUEST") == -1){ url = appendSubRequestArgs(url);}
  if(url.lastIndexOf('#') > -1)
  {
    url= url.substring(0,url.lastIndexOf('#'));
  }
  url += "#" + encodeURIComponent(encodedOptions);
  return url;
}

AjaxAPI.checkIfBlankPage = function(resFrame)
{
   if((resFrame.location.href == "about:blank")
      || (resFrame.location.href.indexOf("blank.html") > 0) 
      || (resFrame.location.href.indexOf("empty.html") > 0))
   {
     return true;
   }
   return false;
}




AjaxAPI.handleIframeResponse = function(respFrameEl)
{
  if(AjaxAPI.navigFrameName == null)
  {
    AjaxAPI.navigFrameEl = respFrameEl;
    AjaxAPI.navigFrameName = respFrameEl.getAttribute("name");
  }

  var respFrame = window.frames[respFrameEl.getAttribute("name")];
  if(AjaxAPI.checkIfBlankPage(respFrame))
  {
    return;
  }
  var reqOptions = respFrameEl.reqOptions;
  respFrame.reqOptions=null;
  if(reqOptions == null)
  {
    var encodedOptions = respFrame.location.hash;
    if((encodedOptions == null) || (encodedOptions.length < 2))
    {
      return;
    }
    encodedOptions = encodedOptions.substring(1);
    if(browser_ie){ encodedOptions = decodeURIComponent(encodedOptions);}
    reqOptions = new AjaxOptions(encodedOptions);
    reqOptions.setV("URL",respFrame.location.href);
  }
  var ifTransport = new Object();
  ifTransport.iframe = respFrame;
  ifTransport.readyState = 4;
  ifTransport.status = 200;
  ifTransport.responseText = "";
  var isTextPlain = false;
  var nodeList = respFrame.document.getElementsByTagName("*");
  var isTextStarted = false;
  var dummyPrefix = "                    ";
  for(var j =0;j < nodeList.length;j++)
  {
    var curNode = nodeList[j];
    if(curNode.nodeName == "PRE")
    {
      var childNodes = curNode.childNodes;
      for(var i = 0; i < childNodes.length; i++)
      {
        if(!isTextStarted)
        {
          if(childNodes[i].nodeValue.indexOf(dummyPrefix) != 0)
          {
            continue;
          }
          isTextStarted = true;
        }
        ifTransport.responseText += childNodes[i].nodeValue;
      }
    }
  }
  if(ifTransport.responseText.indexOf(dummyPrefix) == 0)
  {
    isTextPlain = true;
  }

  if(!isTextPlain)
  {
    ifTransport.responseText = nodeList[0].innerHTML;
  }
  var resp = new AjaxResponse(ifTransport);
  AjaxUtils.handleAjaxResponse(resp,reqOptions);
}




/**
 * Use this method to invoke navigable action(.i.e, actions that need to support Browser Back!!).
 * Experimental!!!
 * @private
 */
AjaxAPI.invokeNavigableAction = function(requestOptions)
{
  var encodedOptions = requestOptions.encode();
  requestOptions.fn("NAVIGFUNCTION_N")(requestOptions);
  AjaxUtils.ignoreNavigFrameLoad= true;
  window.frames["AJAXNAVIG"].location.href = ZDBGlobal.REPORTS_CONTEXT_NAME + CONTEXT_PATH + "/framework/html/blank.html?" + encodedOptions;
}


/**
 * Use this function to submit a form via Ajax. Internally calls {@link #sendRequest} with form's
 * elements values as parameters.
 * <br/><b>Usage :</b> On the form, set this attribute <code>onSubmit = "return AjaxAPI.submit(this);"</code>
 * <br/><b>Note:</b>
 * <ul><li>The various request options for {@link #sendRequest} is got from the form's attributes via 
 *     getAttribute.</li>
 *     <li>In addition <code>validatefunc</code> attribute can be specified to validate the form before 
 *   submitting. For example to use {@link FormHandling.js#validateForm} specify 
 *   <code>validateFunc="validateForm"</code>.</li>
 *</ul>
 * @param {HtmlForm} frm the form to submit. 
 * @return Always returns <i>false</i> as the forms values are sent via XmlHTTP and hence the form need not be submitted.
 * @type boolean 
 */
AjaxAPI.submit = function(frm,options)
{
 var frmOptions = AjaxAPI.getFormAttrsAsOptions(frm);
  try
    {
      if(options)
      {
        frmOptions.update(options);
      }
      if(frmOptions["VALIDATEFUNC"])
      {
        if(!frmOptions.fn("VALIDATEFUNC")(frm))
        {
          return false;
        }
      }
      var reqURL = frm.getAttribute("action");
      if(reqURL.indexOf(location.protocol) != 0 && reqURL.indexOf("/reports/") != 0)
      {
        reqURL = ZDBGlobal.REPORTS_CONTEXT_NAME + reqURL;  
      }
      if(!frmOptions.isTrue("USEIFRAME",false))
      {
      var params = ZRSerialize(frm);        
        frmOptions.update({URL:reqURL,parameters: params});
        var srcview = frmOptions["SRCVIEW"];
        if(srcview == null)
        {
            srcview = DOMUtils.getParentWithAttr(frm, "unique_id");
            if(srcview != null)
            {
                frmOptions.update({SRCVIEW:srcview.getAttribute("unique_id")});
            }
        }
        AjaxAPI.sendRequest(frmOptions);
      }
      else
      {
        if(frmOptions["METHOD"].toLowerCase() == "post")
        {
          var csrfEl = frm.querySelector("#ZDBCSRFToken"); //No I18N
          if(csrfEl)
          {
            csrfEl.setAttribute("value",getCookie("CSRF_TOKEN"));
          }
          else
          {
            var inputEl = document.createElement("input");
            inputEl.setAttribute("ID","ZDBCSRFToken");
            inputEl.setAttribute("type","hidden");
            inputEl.setAttribute("name","ZDB_CSRF_TOKEN");
            inputEl.setAttribute("value",getCookie("CSRF_TOKEN"));
            frm.insertBefore(inputEl,frm.firstChild);
          }
        }
        frm.target=AjaxAPI.navigFrameName;       
        var ajaxOpt = new AjaxOptions(frmOptions);
        if(reqURL.indexOf("SUBREQUEST") == -1 && !frmOptions["NOSUBREQPARAM"])
        { 
          reqURL = appendParamsToUrl(reqURL,"SUBREQUEST=true&_ZVER_="+ZDBGlobal._ZVER_);// No I18n
        }
        ajaxOpt.update({URL:reqURL,ONSUCCESSFUNC:'AjaxAPI.iframeSubmitResponse',ONFAILUREFUNC:'AjaxAPI.iframeSubmitFailure',FORMNAME:frm.name});        //No i18N
        frm.action=reqURL;
        
        /**** Progress bar handling start 
         * It expects three parameters. 
         * PROGRESSIND - Whether to handle progress/not
         * PROGPERCDIV - Progress percentage DIV - Texual percentage value will be shown here. If not sent, will be ignored
         * PROGRESSDIV - Progress indicator. You can style as per your wish. Except that, the width should be 0% at starting. Will be 100% after full load.
         * 
         * Sample HTML code
         * 
         * <div style="position:relative;">
         *    <div id="PROGRESSDIV" style="height:100%;position:absolute;background-color:green;opacity:0.4;left:0px;"></div>
       *      Step 2<%--No I18N--%> 
       *    <span id="PROGPERCDIV"></span>
         *  </div>
         * 
         * *******/
        if(frmOptions.PROGRESSIND === "true")
      {
          AjaxAPI.fileUploadHandler(frm,ajaxOpt);
      }
        else
        { 
          StatusMsgAPI.showOperationStatus(StatusMsgAPI.OPSTATUS.STARTED,ajaxOpt["OPSTATUSID"],ajaxOpt["LOADINGTYPE"], ajaxOpt['PARENT']);
          AjaxAPI.navigFrameEl.reqOptions=ajaxOpt;
          frm.submit();
        }
        
        DOMUtils.enableElements(frmOptions["DISABLEELS"],false);
      }
    }
  catch(e)
    {
      StatusMsgAPI.showMsg("Script Error Occurred : " + LangUtils.getErrorMsg(e)
                           ,false,true,frmOptions["STATUSMSGID"]);
      if(frmOptions["HANDLEEXEC"]){
    	  frmOptions["HANDLEEXEC"]();
      }
    }
  return false;
}

/**** Progress bar handling start *******/
AjaxAPI.fileUploadHandler = function(frm,ajaxOpt)
{
  var oReq = new XMLHttpRequest();
    oReq.open("post", frm.action, true);

    var resp = new AjaxResponse(oReq);
    oReq.onreadystatechange = function(){AjaxUtils.handleAjaxResponse(resp,ajaxOpt);};

    if(ajaxOpt.STATUSFUNC)
    {
      ajaxOpt.fn("STATUSFUNC")(resp,ajaxOpt);//No I18N
    }
    
  var upload = oReq.upload;
    upload.downloadStartTime = new Date().getTime();
    upload.currentStart = upload.downloadStartTime;
    upload.currentProgress = 0;
    upload.startData = 0;
    
    //Event to calculate the load when it can be computable
    upload.addEventListener("progress", function(event){AjaxAPI.updateProgress(event,ajaxOpt);});
    
    //Event to be called after request sent
    upload.addEventListener("load", function(event){AjaxAPI.completeProgress(event,ajaxOpt);});
    
    if(ajaxOpt.INITFUNC){
    	var initfunc = ajaxOpt.fn("INITFUNC"); //No I18N
    	initfunc();
    }
    oReq.send(new FormData(frm));
}

AjaxAPI.completeProgress = function(event, ajaxOpt)
{ 
	if(ajaxOpt.COMPLETEFUNC){
		var completeFunc = ajaxOpt.fn("COMPLETEFUNC"); //No I18N
		completeFunc();
	} 
}

AjaxAPI.updateProgress = function(event,ajaxOpt)
{
	var totalSizeSet = false;
	if (event.lengthComputable) {
	   var dataLoaded = event.loaded;
	   var totalSize = event.total;
	   if(!totalSizeSet && ajaxOpt.TOTALDIV){
		   jQuery("#"+ajaxOpt.TOTALDIV).html(" / "+AjaxAPI.changeFileSizeUnit(totalSize));
		   totalSizeSet = true;
	   }
	   var progress = Math.ceil( (dataLoaded / totalSize) * 100 );
	   if(ajaxOpt.PROGPERCDIV)
	   {
		   jQuery("#"+ajaxOpt.PROGPERCDIV).html(progress + "%");//No I18N
	   }
	   if(ajaxOpt.PROGLOADEDDIV){
		   jQuery("#"+ajaxOpt.PROGLOADEDDIV).html(AjaxAPI.changeFileSizeUnit(dataLoaded));//No I18N
	   }
       jQuery("#"+ajaxOpt.PROGRESSDIV).css("width",progress + '%');//No I18N
  }
}

AjaxAPI.changeFileSizeUnit =  function(size){
	if(!isNaN(size)){
		if(size>=1024 && size<1048576){
			size =  size/1024;
			size = size.toFixed(2);
			size =  size +"KB"; //NO I18n
		}
		else if(size>1048576){
			size =  size/1048576;
			size = size.toFixed(2);
			size =  size +"MB"; //NO I18n
		}
		else{
			size = size.toFixed(2);
			size = size +"Bytes"; //NO I18n
		}
	}
	return size;
}

/**** Progress bar handling END *******/

AjaxAPI.getFormAttrsAsOptions = function(frm)
{
  var attrList = frm.attributes;
  var frmOptions = new AjaxOptions({submittedFrm:frm.name});
  for(var i = 0; i < attrList.length; i++)
  {
    frmOptions.setV(attrList.item(i).nodeName, attrList.item(i).nodeValue);
  }
  return frmOptions;
}

AjaxAPI.iframeSubmitResponse = function(response,reqOptions)
{
  var frm = DOMUtils.getForm(reqOptions['FORMNAME']);
  var frmOptions = AjaxAPI.getFormAttrsAsOptions(frm);
  StatusMsgAPI.showOperationStatus(response.opStatus,reqOptions["OPSTATUSID"],reqOptions["LOADINGTYPE"], reqOptions['PARENT']);  
  if(frmOptions["ONSUCCESSFUNC"] != null)
  {
    frmOptions.fn("ONSUCCESSFUNC")(response,frmOptions);
  }
}

AjaxAPI.iframeSubmitFailure = function(response,reqOptions)
{
  var frm =  DOMUtils.getForm(reqOptions['FORMNAME']);
  var frmOptions = AjaxAPI.getFormAttrsAsOptions(frm);
  AjaxAPI.showRespMsg(response,frmOptions);
}

/**
 * Dummy Function. Does Nothing.
 * @param {AjaxResponse} response
 * @param {AjaxOptions} requestOptions
 * @return {@link StatusMsgAPI.OPSTATUS#FINISHED}
 * @type StatusMsgAPI.OPSTATUS
 */
AjaxAPI.dummyFunction = function(response,requestOptions)
{
  return StatusMsgAPI.OPSTATUS.FINISHED;
}


/**
 * Refreshes the view specified in the <code>SRCVIEW</code> option specified in the requestOptions
 * @param {AjaxResponse} response
 * @param {AjaxOptions} requestOptions
 * @return {@link StatusMsgAPI.OPSTATUS#INPROGRESS}
 * @type StatusMsgAPI.OPSTATUS
 */
AjaxAPI.refreshView = function(response,requestOptions)
{
  requestOptions.update({VIEWTOREFRESH_RN:requestOptions.v("SRCVIEW"),NAVIGABLE_RN:false});
  ViewAPI.refreshView(requestOptions);
  return StatusMsgAPI.OPSTATUS.INPROGRESS;
}


/**
 * Sets the ajax attributes on html form. For more details about the options look 
 * at the table ACAjaxForm in Client Framework.
 * @param {Object} ajaxOptions.
 */
AjaxAPI.setAjaxAttributes = function(ajaxOptions)
{
  ajaxOptions = new AjaxOptions(ajaxOptions);
  var frm = ajaxOptions.form("FORMNAME");
  for(var i in ajaxOptions)
  {
    if((typeof ajaxOptions[i]) != "string")
    {
      continue;
    }
    existValue = ajaxOptions[i];
    frm.setAttribute(i.toLowerCase(),existValue);
  }
  if(!frm.onsubmit)
  {
    frm.onsubmit = function(){return AjaxAPI.submit(this);}; 
  }
  frm.requestOptions = ajaxOptions;
}


/**
 * Shows the response status and success/failure messages.
 * @param {AjaxResponse} response
 * @param {AjaxOptions} requestOptions 
 */
AjaxAPI.showRespMsg = function(response,reqOptions)
{
  var opStatus = response.opStatus;
  var dontShowOperationStatus = reqOptions["DONTSHOWOPSTATUS"];
  if(!dontShowOperationStatus){
	  StatusMsgAPI.showOperationStatus(opStatus,reqOptions["OPSTATUSID"],reqOptions["LOADINGTYPE"], reqOptions['PARENT']);
  }
	  if(opStatus == StatusMsgAPI.OPSTATUS.STARTED)
  {
    StatusMsgAPI.closeMsgImmediately(reqOptions["STATUSMSGID"]);
    return;
  }
  else if(opStatus == StatusMsgAPI.OPSTATUS.FAILED_SCRIPT)
  {
    throw response["EXCEP"];
  }
  else if(opStatus == StatusMsgAPI.OPSTATUS.FAILED)
  {
    if(AjaxAPI.storeLastFailedResp)
    {
      AjaxAPI.lastFailedContent = response.responseText;
    }
    if(reqOptions["ONFAILUREFUNC"] != null)
    {
      var handled = reqOptions.fn("ONFAILUREFUNC")(response,reqOptions);
      if((AjaxAPI.ONFAILUREFUNCHANDLED) || (handled))
      {
        return;
      }
    }
    var respMsg = response.getResponsePart("STATUS_MESSAGE");
    if(respMsg == null)
    {
      StatusMsgAPI.showErrorMsg(response.getOnlyHtml());
    }
    else
    {
      StatusMsgAPI.showMsg(respMsg,(opStatus == StatusMsgAPI.OPSTATUS.FINISHED),true,reqOptions["STATUSMSGID"]);
    }
  }
  else if(opStatus == StatusMsgAPI.OPSTATUS.FINISHED)
  {
    var respMsg = response.getResponsePart("STATUS_MESSAGE");
    if(respMsg == null)
    {
      var prevResp = reqOptions["PREVRESPONSE_RN"];
      if(prevResp != null)
      {
        respMsg = prevResp.getResponsePart("STATUS_MESSAGE");
      }
    }
    if(respMsg != null)
    {
      StatusMsgAPI.showMsg(respMsg,(opStatus == StatusMsgAPI.OPSTATUS.FINISHED),true,reqOptions["STATUSMSGID"]);
    }
  }
  else if(opStatus == StatusMsgAPI.OPSTATUS.INPROGRESS)
  {
    reqOptions["PREVRESPONSE_RN"] = response;
  }

  if((!AjaxAPI.checkAttrForInvokeScript) || (reqOptions["INVOKESCRIPTS"]))
  {
    response.invokeScripts();
  }
  
}




/**
 *@class
 * The class used to hold the ajax request options. It has some utility methods over
 * the default javascript object.
 * <br><b>Usage:</b>
 * <br>&nbsp;&nbsp;<code>var ajOpts = new AjaxOptions({URL:"test.cc",METHOD:"GET"});</code>
 * <br><b>Note:</b> All options will be automatically converted to  UpperCase. 
 *
 */

function AjaxOptions(options)
{

  /**
   * Used to return the value for the option. Throws Error if option is not present and the default
   * value is also not given.
   * @param {String} optName.
   * @param {Object} defaultValue.
   * @type Object
   * @throws Error
   */
  this.v = function(optName,defaultValue)
    {
      var value = this[optName.toUpperCase()];
      if(value == null)
      {
        value = defaultValue;
      }
      if(value == null) 
      {
        throw new Error("OptionNotPresent:" + optName + " is not present");
      }
      return value;
    }

  
  /**
   * Used to return the value in uppercase for the option. Throws Error if option is not present.
   * @param {String} optName.
   * @type String
   * @throws Error
   */
  this.vu = function(optName)
    {
      return this.v(optName).toUpperCase();
    }


  /**
   * Used to return the boolean value for the option. Throws Error if option is not present.
   * <br/> Handles cases where the value could be a string "true".
   * @param {String} optName.
   * @type boolean
   * @throws Error
   */
  this.isTrue = function(optName,defaultValue)
    {
      var value =  this.v(optName,defaultValue);
      return ((value == true) || (value == "true"));
    }


  /**
   * Used to return the function associated for the option. 
   * <br>Throws Error if 
   *  <ul><li> Option itself is not configured.</li><li>Corresponding function is not present</li></ul>
   * <br/> Handles cases where the value configured could be of type
   * <ul><li>String representing th function name</li><li>Reference to a function</li>
   * @param {String} optName.
   * @type Function
   * @throws Error
   */
  this.fn = function(optName)
    {
      var value = this.v(optName);
  
      if(value instanceof Function)
      {
        return value;
      }
      
      var fun = (new Function('return ' + value))();
      if(!fun)
      { 
        throw new Error("FunctionNotPresent: " + value + " is not declared in current window");   
      }
      if(!(fun instanceof Function))
      {
        throw new Error("IllegalOption : " + value + " is not a function.");
      }
      return fun;
    }


  /**
   * Used to return the function associated for the option. Throws Error if option is not present.
   * <br/> Handles cases where the value configured could be of type
   * <ul><li>String representing th function name</li><li>Reference to a function</li>
   * @param {String} optName.
   * @type Function
   * @throws Error
   */
  this.form = function(optName)
    {
      var frm = DOMUtils.getForm(this.v(optName));
      if(frm == null)
      {
        throw new Error(" FormNotFound : " + this.v(optName) + " not found");  
      }
      return frm;
    }

  
  /**
   * Sets the value for an option. Stores the option in uppercase.
   * @param {String} optName.
   * @param {Object} optValue.
   */
  this.setV = function(optName,optValue)
    {
      this[optName.toUpperCase()] = optValue;
    }

  
  /**
   * Sets the value for an option. Stores the option in uppercase.
   * @param {String} optName.
   * @param {Object} optValue.
   */
  this.update = function(newOptions)
    {
      for (property in newOptions) 
      {
        var val = newOptions[property];
        this[property.toUpperCase()] = val;
      }
    }


  /**
   * Encodes the option into a query parameter. Only simple type(String/numeric) options are encoded.
   * Any parameters ending with _RN (implying Remove for Navigation) will not be added. Also URL parameter
   * will not be added.
   * @return A query string.
   * @type String.
   */
  this.encode = function()
    {
      var encodedStr ="";
      var defaults = new AjaxOptions();
      for(var i in this)
      {    
        if(this[i] instanceof Object)
        {continue;}
        if(defaults[i] == this[i])
        { continue;}
        
        if((i == "URL") || (i =="METHOD"))
        { continue;}
        var index = i.lastIndexOf("_RN");
        if(index == i.length - "_RN".length)
        {
          continue;
        }
        
        var name = (i == "ONSUCCESSFUNC")? "_OS":i;
        encodedStr += name + "=" + this[i] + "&";
      }
      if(encodedStr.length > 1)
      {
        encodedStr = encodedStr.substring(0,encodedStr.length -1);
      }
      return encodedStr;
    }

  this.update({METHOD:'POST',ASYNCHRONOUS:true,STATUSFUNC:'AjaxAPI.showRespMsg',ONFAILUREFUNC:AjaxAPI.onFailureFunc});
  if((typeof options) == "string")
  {
    options = parseRequestParams(options);
    for(var i in options)
    {
      options[i]= options[i][0];
    }
    if(options["_OS"] != null){options["ONSUCCESSFUNC"] = options["_OS"]}
  }
  this.update(options);
}

/**
 * If the "options" param is not of type AjaxOptions then,
 *  it creates a new AjaxOption and returns otherwise the same
 *  "options" is returned
 * @param {Map/AjaxOptions} options
 * @type AjaxOptions
 */
AjaxOptions.getAsAjaxOptions = function(options)
{
  if(!(options instanceof AjaxOptions))
  {
    options = new AjaxOptions(options);
  }  
  return options;
}

/**
 * @class
 * The Class that is used to represent the xml http response.
 */
function AjaxResponse(transportUsed)
{
  /**
   * Holds the response text.
   * @type String
   */
  this.responseText = null;

  /**
   *@private
   */
  this.htmlContent = null;

  /**
   *@private
   */
  this.scripts = null;

  /**
   * Holds the transport used for this request. Generally it is XMLHttpRequest
   * got from {@link AjaxAPI#getXMLHttpRequest}
   */
  this.transport = transportUsed;

  /**
   * Holds the current status of the response.
   * @type StatusMsgAPI.OPSTATUS
   */
  this.opStatus = StatusMsgAPI.OPSTATUS.STARTED;

  
  /**
   * Checks if http request protocol status is success.This is <b>different</b>
   * from overall success which would be captured in {@link StatusMsgAPI.OPSTATUS}!!.
   */
  this.isHttpReqSuccess = function() 
    {
      try
      {
        return this.transport.status == undefined
        || this.transport.status == 0 
        || (this.transport.status >= 200 && this.transport.status < 300);
      }
      catch(e)
      {
        return false;
      }
    }

  this.getResponseStatus = function()
    {
      try
      {
        return this.transport.status;
      }
      catch(e)
      {
        return 999999;
      }
    }

  this.getResponseStatusText = function()
    {
      try
      {
        return this.transport.statusText;
      }
      catch(e)
      {
        return null;
      }
    }
  
  /**
   * Gets a part of the response. That part should be present in a div with
   * a custom attribute <code>type</code> set to the responsePart specified.
   * @return A String if the response contains that div or null.
   * @type String
   */  
  this.getResponsePart = function(responsePart)
    {
      return AjaxUtils.extractFromResponse(this.responseText,responsePart);
    }


  /**
   * Gets a response parameters. These should have been passed as a map to 
   * the server side api  WebViewAPI.sendResponse.
   * @type Map
   */    
  this.getResponseParams = function()
    {
      var paramString = AjaxUtils.extractFromResponse(this.responseText,"RESPONSE_PARAMS");
      return (paramString != null)?eval(paramString):{};
    }


  /**
   * Returns the response text stripped of script tags. Incase of adding the 
   * response to the document, then use this function rather than directly adding
   * {@link #responseText}.
   * @type String
   */
  this.getOnlyHtml = function()
    {
      this.splitResponse();
      return this.htmlContent;
    }


  /**
   * Returns the script tags present in the responseText as an array.
   * see {@link #invokeScripts} for more details.
   * @type Strings[]
   */
  this.getScripts = function()
    {
      this.splitResponse();
      return this.scripts;
    }

  /**
   * Invokes the script tags. This is automatically called at the end of the 
   * processing by the function configured in STATUSFUN. To add to the list of
   * scripts to be called call {@link #getScripts} and append to the array.
   * See {@link AjaxUtils#invokeScripts} for more details.
   */
  this.invokeScripts = function()
    {
      if(!this["alreadyInvoked"])
      {
        AjaxUtils.invokeScripts(this.getScripts());
        this["alreadyInvoked"] = true;
      }
    }

  /**
   *@private
   */
  this.splitResponse = function()
    {
      if(this.htmlContent !=null)
      {
        return;
      }
      var split = AjaxUtils.separateScriptAndHtml(this.responseText);
      this.htmlContent = split["html"];
      this.scripts = split["scripts"];
    }
}


/**
 *@class
 * The <b>Static</b> utility class containing methods related to Ajax.
 * Access all functions using AjaxUtils.<functionname>
 */
var AjaxUtils = new function(){}




/**
 *@private
 */
AjaxUtils.ignoreNavigFrameLoad = false;

/**
 *@private
 */
AjaxUtils.handleNavigFrameLoad = function()
{
  var queryStr = window.frames["AJAXNAVIG"].location.search;
  window.status = "handle NavigFrameLoad called. " + queryStr + " " + AjaxUtils.ignoreNavigFrameLoad;

  if(AjaxUtils.ignoreNavigFrameLoad) 
  {
    AjaxUtils.ignoreNavigFrameLoad = false;
     return;
  }

  if((queryStr == null) || (queryStr == ""))
  {
    return;
  }
  var requestOptions = new AjaxOptions(queryStr);
  requestOptions.fn("NAVIGFUN_N")(requestOptions);
}


/**
 * Utility method that can be used invoke the scripts tag in the. See {@link #invokeScripts}
 * for more details.
 * @param {String} text Should be a html text.
 * @see #separateScriptAndHtml
 */
AjaxUtils.scriptFragment = '(?:<script.*?>)((\n|.)*?)(?:<\/script>)';

AjaxUtils.invokeScriptsInHtml = function(html)
{
  var match    = new RegExp(AjaxUtils.scriptFragment, 'img');
  var scriptTags  = html.match(match);
  AjaxUtils.invokeScripts(scriptTags);
}

/**
 * A utility method used to separate the script and html parts!!.
 * @return An object containing properties "html" containing a string stripped of script tags 
 *  and "script" containing an array of script tags that were present.
 * @param {String} text. 
 * @type Object
 */
AjaxUtils.separateScriptAndHtml = function(htmlSnippet)
{
  var sepHash = new Object();
  var match    = new RegExp(AjaxUtils.scriptFragment, 'img');
  sepHash["html"] = htmlSnippet.replace(match, '');
  match    = new RegExp(AjaxUtils.scriptFragment, 'img');
  sepHash["scripts"] = htmlSnippet.match(match);
  return sepHash;
}


/**
 * Invokes the scripts present in the array. 
 * It handles scripts having "src" attribute also. The scripts will be immediately
 * executed but will be executed asynchronouly. 
 * @see #separateScriptAndHtml
 * @param {Strings[]} scriptTags contain script tags 
 */
AjaxUtils.invokeScripts = function(scriptTags)
{
  if(scriptTags == null){return;}
  var div = document.createElement("div");
  var scriptStr = "";
  for (var i = 0; i < scriptTags.length; i++) {
    scriptStr += "\n" + scriptTags[i];
  }
  var match    = new RegExp("<script>", 'img');
  scriptStr = scriptStr.replace(match,"<SCRIPT>");
  match    = new RegExp("<\/script>", 'img');
  scriptStr = scriptStr.replace(match,"</SCRIPT>");
  div.innerHTML = "<html><body>" + scriptStr + "</body></html>";

  
  scriptTags = div.getElementsByTagName("SCRIPT");
  for (var i = 0; i < scriptTags.length; i++) {
    AjaxUtils.scheduleScript(scriptTags[i].text,scriptTags[i].src);
  }
}


/**
 * Schedules the script to be invoked. Note : The script is invoked <b>immediately</b> but not
 * synchronously!!.
 * @param {String} script The script to be invoked.Null if script file is passed.
 * @param {String} scriptFile The script file to load. As in <code>src</code> attribute of <code>script</code> tag. Null if script is passed.
 */
AjaxUtils.scheduleScript = function(script,scriptFile){
  var scriptTag = document.createElement("SCRIPT");
  if ((scriptFile != null) && (scriptFile != ""))
  { scriptTag.src = scriptFile;}

  scriptTag.text = script;
  if (!document.getElementsByTagName("HEAD")[0]) {
    document.createElement("HEAD").appendChild(scriptTag)
      } else {
        document.getElementsByTagName("HEAD")[0].appendChild(scriptTag);
      }
}


/**
 * @private
 */
AjaxUtils.handleAjaxResponse = function(ajaxResp,requestOptions)
{
  if(ajaxResp.transport.readyState != 4 )
  {
    if(requestOptions["ONABORTFUNC"])
    {    
      var abort = requestOptions.fn("ONABORTFUNC")(ajaxResp,requestOptions); //No I18N
      if(abort)
      {
        StatusMsgAPI.showOperationStatus(StatusMsgAPI.OPSTATUS.FINISHED,requestOptions["OPSTATUSID"],requestOptions["LOADINGTYPE"], requestOptions['PARENT']);
        ajaxResp.transport.abort();
        return;
      }
    }
    
    /**** Progress bar -with updates from server- handling start *******/
    if(requestOptions["MONITORFUNC"]){
    	requestOptions.fn("MONITORFUNC")(ajaxResp,requestOptions); //No I18N
    }
    /**** Progress bar handling End *******/
    
    return;
  }
  ajaxResp.responseText = ajaxResp.transport.responseText;
  ajaxResp.opStatus = StatusMsgAPI.OPSTATUS.INPROGRESS;
  if(requestOptions["ORIGOPTIONS"])
  {
    requestOptions = requestOptions["ORIGOPTIONS"];
    closeDialog();
  }
  DOMUtils.enableElements(requestOptions["DISABLEELS"],true);
  ajaxResp.opStatus = AjaxUtils.internalHandleAjaxResponse(ajaxResp,requestOptions);
  if(requestOptions["STATUSFUNC"])
  {    
    requestOptions.fn("STATUSFUNC")(ajaxResp,requestOptions);
  }
}


/**
 *@private
 */ 
AjaxUtils.internalHandleAjaxResponse = function(response,requestOptions)
{
  var index = 0;
  if(AjaxAPI.statusSearchIndex){ index = response.responseText.length - AjaxAPI.statusSearchIndex;}
  if(index < 0)
  {
    index = 0;
  }
  if(response.responseText.indexOf("__ERROR__LOGIN",index) > -1)
  {
    AjaxUtils.handleReLogin(response,requestOptions);
    return StatusMsgAPI.OPSTATUS.INPROGRESS_LOGIN;
  }
  else if((response.responseText.indexOf("__ERROR__",index) > -1)
          || !response.isHttpReqSuccess())
  {
    return StatusMsgAPI.OPSTATUS.FAILED;
  }
  var opStatus = StatusMsgAPI.OPSTATUS.FINISHED;
  if(requestOptions["ONSUCCESSFUNC"])
  {
    try
    {
       opStatus = requestOptions.fn("ONSUCCESSFUNC")(response,requestOptions);
       if(!opStatus){opStatus = StatusMsgAPI.OPSTATUS.FINISHED;}
    }
    catch(ex)
    {
      response["EXCEP"] = ex;
      return StatusMsgAPI.OPSTATUS.FAILED_SCRIPT;
    }
  }
  return opStatus;
}


/**
 * @private
 */
AjaxUtils.handleReLogin = function(response,requestOptions)
{
  showDialog(response.responseText,"title=ReLogin(Session Timed Out),modal=yes");
  var frm = DOMUtils.getFirstMatchingElement(oDialog.getElementsByTagName("form"),null,"name","login");
  if(frm == null)
  {
    throw new Error("Login Page should contain a form with name as \"login\"");
  }
  frm["REQUESTOPTIONS"] = requestOptions;
  frm.onsubmit = function(){AjaxUtils.submitReLoginFrm(frm);return false;};
  return;
}



/**
 * @private
 */

AjaxUtils.updateObject = function(destination, source) {
  if(source == null){return;}
  for (property in source) {
    destination[property] = source[property];
  }
  return destination;
}


/**
 * @private
 */
AjaxUtils.extractFromResponse = function(responseText,responsePart)
{
  var div = document.createElement("div");
  div.innerHTML = responseText;
  var respPart = DOMUtils.getFirstMatchingElement(div.getElementsByTagName("div"),null,"part",responsePart);
  return (respPart != null)? respPart.innerHTML: null;
}

/** Use this method to load resources like js and css **/
AjaxUtils.sendResourceGetRequest = function(url,requestOptions){
  var reqObj = AjaxAPI.getXMLHttpRequest();
    
  if (BROWSER.IE && BROWSER.VERSION && BROWSER.VERSION.indexOf("9") === 0 && 
     'XDomainRequest' in window && window.XDomainRequest !== null) {
    //IE9
    reqObj = new XDomainRequest();
  } else {
    // For all other browsers
    reqObj = new XMLHttpRequest();
  }
  reqObj.open("GET", url);
  reqObj.send();
  var resp = new AjaxResponse(reqObj);
  resp.url = url;
  reqObj.onload = function(){
    // if(resp.transport.readyState == XMLHttpRequest.DONE){
      var func = requestOptions["ONSUCCESSFUNC"];
      func.apply(requestOptions["OBJCTX"],[resp, requestOptions])
    // }
  };
}

/*
 * Use this method to set the default params that needs to be sent when authentication will not be available and to skip the authentication
 * by validating the params that are being sent.
 */

AjaxUtils.addAdditionalParamsForAuthenticationSkip = function(url){
    //In slideshow cases all the authentication will be done based on the slideID and slideKey
    if(ZDBGlobal.IS_SLIDESHOW){
        var slideShowParams = '';
        slideShowParams += "SLIDEID="+ZDBGlobal.SLIDESHOW_ID+"&SLIDEKEY="+ZDBGlobal.SLIDESHOW_KEY;
        url = appendParamsToUrl(url, slideShowParams);
    }
    return url;
}

function executeFunctionAsString(functionName, context) {
//    var namespaces = functionName.split(".");
//    var func = namespaces.pop();
//    for (var i = 0; i < namespaces.length; i++) {
//        context = context[namespaces[i]];
//    }
    if (typeof functionName === "function")
    {
        if ((arguments[2]).callee)
        {
            var index = isWindow(arguments[2]);
            if (index == -1)
            {
                return functionName.apply(context, Array.prototype.slice.call(arguments[2]).splice(1, (arguments[2].length - 1)));
            } else
            {
                return functionName.apply(context, Array.prototype.slice.call(arguments[2]).splice(index + 1, (arguments[2].length - 1)));
            }
        } else if (arguments.length > 1)
        {
            return functionName.apply(context, Array.prototype.slice.call(arguments).splice(2, (arguments[2].length - 1)));
        } else
        {
            return functionName.apply(context);
        }
    }
}
function isWindow(params)
{
    for (var i = 0; i < params.length; ++i) {
        if (params[i].toString() == "[object Window]") {
            return i;
        }
    }
    return -1;
}
