(function (win, global) {

    "use strict";
    var garbageBin, Console = win.console, helperContainer = null;

    if (!global) {
        throw new ReferenceError("$ZC not found");
    }

    var defaultOptions = {
        "exporting": {
            "url": "",
            "filename": "chart",
            "type": "png",
            "scale": 1
        }
    };

    var namespace = {
        prefix: {
            svg: "http://www.w3.org/2000/svg",
            xhtml: "http://www.w3.org/1999/xhtml"
        }
    }, exporting = false;

    /**
     *
     * @param objectLikeArray
     * @returns {Array}
     */
    function toArray(objectLikeArray) {
        return Array.prototype.slice.call(objectLikeArray);
    }


    /**
     *
     * @param {Object} chartObj
     * @param {Object} options
     */

    var exportChart = function (chartObj, options) {

        // basic validation
        if (!defined(chartObj)) {
            throw new ReferenceError("Chart Object not found");
        } else if (!defined(chartObj.svg) || !defined(chartObj.systemConf)) {
            throw new ReferenceError("Invalid Chart Object");
        } else if (!chartObj.chartRendered) {
            return Console.warn("Chart rendering is not completed");
        }

        var dimension = chartObj.container.node().getBoundingClientRect();
        options = merge(defaultOptions.exporting, options);

        if(options.client === true) {
            clientExport(options, chartObj, dimension)
        } else {
            serverExport(options, chartObj, dimension)
        }
    };

    function serializeChartSvg(chart) {

        if (!helperContainer) {
            helperContainer = document.querySelector(".__baseZC_Container__");
        }

        var chartSvg = chart.container.node(),
            patternSVG = helperContainer.querySelector("svg.svgPattern"),
            dimension = chartSvg.getBoundingClientRect(),
            svgPrefix = namespace.prefix.svg,
            xhtmlPrefix = namespace.prefix.xhtml;

        // clone the chart

        var chartClone = chartSvg.cloneNode(true);
        var patternClone = patternSVG.cloneNode(true); // has the  xmlns attribute. :)

        //set xmlns attribute to all necessary the elements.

        chartClone.setAttribute("xmlns", xhtmlPrefix);
        each(chartClone.querySelectorAll("svg"), function(svg) {
            svg.setAttribute("xmlns", svgPrefix);
        });

        each(chartClone.querySelectorAll("div, ul, li"), function(el) {
            el.style.setProperty("box-sizing", "border-box", "important");
        });


        // convert legend filter box svg background shapes to base64
        var legends = chartClone.querySelectorAll(".legendfilterbox");
        each(legends, function(legend) {
            var backgroundSVG = legend.style.backgroundImage.match(/<svg.+\/svg>/ig);
            if(backgroundSVG) {
                backgroundSVG = backgroundSVG[0].replace(/["']/ig, "'");
                legend.style.backgroundImage = "url(" + buildSvgImageUrl(backgroundSVG) + ")";
            }
        });

        // create the container svg
        var svg = document.createElementNS(svgPrefix, "svg");
        svg.setAttribute("xmlns", svgPrefix);
        svg.setAttribute("width", dimension.width);
        svg.setAttribute("height", dimension.height);

        var foreignObject = document.createElementNS(svgPrefix, 'foreignObject');
        foreignObject.setAttribute("width", "100%");
        foreignObject.setAttribute("height", "100%");
        svg.appendChild(foreignObject);

        // append the pattern element directly to the chart container
        chartClone.appendChild(patternClone);

        // append the chart container in the foreignObject
        foreignObject.appendChild(chartClone);

        // if there any images convert to data uri
        convertImagesToURI(svg);
        copyCanvas(chartSvg, svg);

        return svg;
    }


    /**
     *
     * @param img
     * @param element
     * @param callback
     */
    function getDataUri(img, element, callback) {
        var toURI = function(image) {
            var canvas = document.createElement('canvas');
            canvas.width = image.naturalWidth || image.getAttribute("width");
            canvas.height = image.naturalHeight || image.getAttribute("height");
            canvas.getContext('2d').drawImage(image, 0, 0);
            return canvas.toDataURL('image/png', 1)
        };

        if(img instanceof HTMLImageElement || img instanceof SVGImageElement) {
            return toURI(img)
        } else {
            var image = new Image();
            image.onload = function () {
                callback(toURI(this), element);
            };
            image.src = img;
        }
    }

    /**
     *
     * @param source
     * @param destination
     */

    function copyCanvas(source, destination) {

        var sourceCanvases = toArray(source.querySelectorAll("canvas")),
            referenceCanvas = toArray(destination.querySelectorAll("canvas"));

        sourceCanvases.forEach(function(canvas, i) {
            var image = document.createElement("div");
            // copy all style
            image.style = canvas.getAttribute("style");
            image.style.backgroundImage = "url(" + canvas.toDataURL("image/png", 1) + ")";
            referenceCanvas[i].parentNode.insertBefore(image, referenceCanvas[i]);
        });
    }


    /**
     *
     * @param {HTMLElement} parentElement
     */
    function convertImagesToURI(parentElement) {
        var images = toArray(parentElement.querySelectorAll("img, image")), svgImagesCompleted = true;

        images.forEach(function(img) {
            svgImagesCompleted = false;
            img.setAttributeNS('http://www.w3.org/1999/xlink', 'href', getDataUri(img));
        });

        // handle all background images.

        var backgroundImages = toArray(parentElement.querySelectorAll("[style*='background-image']")),
            count = 0;

        backgroundImages = backgroundImages.filter(function(element) {
            return /url\(/i.test(element.style.backgroundImage)
        });

        backgroundImages.forEach(function(element) {
            exporting = true; // async
            var url = element.style.backgroundImage.slice(4, -1).replace(/"/g, "");
            getDataUri(url, element, function(uri, el) {
                count++;
                el.style.backgroundImage = "url(" + uri + ")";
                exporting = count !== backgroundImages.length;
            })
        })
    }

    /**
     *
     * @param {String} source base64 string
     * @param {Object} settings user export settings.
     */
    function download(source, settings) {
        var downloadName = (settings.name || "chart") + "." + settings.type;
        var anchor = createElement("a", {
            "download": downloadName,
            "href": source
        }, null, helperContainer);
        anchor.click();
        removeElement(anchor);
    }

    /**
     *
     * @param pixels
     * @param DPI
     * @returns {number}
     */
    function pixelToMillimeter(pixels, DPI) {
        return ( pixels * 25.4 ) / DPI
    }

    /**
     *
     * Using this function over window.createObjectURL, because for security reason
     * browser does not allow exporting a Canvas if a foreignObject has drawn on it.
     *
     * stackoverflow.com/questions/39148582/
     *
     * @param {string} svg html string
     * @returns {string}
     */
    function buildSvgImageUrl(svg) {
        var b64 = win.btoa(win.unescape(win.encodeURIComponent(svg)));
        return "data:image/svg+xml;base64," + b64;
    }

    /**
     *
     * @param {Object} settings user export settings.
     * @param {string} source svg string
     * @param {ClientRect} dimension bounding box of the main container.
     * @param {Function} callback
     */

    function handleImageExport(settings, source, dimension, callback) {

        var devicePixelRation = Math.max(settings.scale, win.devicePixelRatio, 2);

        // create temp high dpi canvas
        var canvas = createElement("canvas");
        var context = canvas.getContext('2d');
        canvas.width = dimension.width * devicePixelRation;
        canvas.height = dimension.height * devicePixelRation;
        context.scale(devicePixelRation, devicePixelRation);

        var img = new Image();
        img.crossOrigin = "Anonymous";

        img.onload = function() {
            context.drawImage(img, 0, 0);
            // document.body.appendChild(this);
            var base64Image = canvas.toDataURL("image/" + settings.type, 1); // 1 highest quality. default is 0.92
            if(defined(callback) && typeof callback === "function") {
                return callback.call(this, base64Image, devicePixelRation);
            }
            download(base64Image, settings);
        };
        img.src = buildSvgImageUrl(source.outerHTML);
    }

    /**
     *
     * @param settings
     * @param source
     */
    function handleSvgExport(settings, source) {

        var serializer = new XMLSerializer();
        source = serializer.serializeToString(source);
        //add xml declaration
        source = '<?xml version="1.0" standalone="no"?>\r\n' + source;

        //convert svg source to URI data scheme.
        var url = "data:image/svg+xml;charset=utf-8,"+encodeURIComponent(source);
        download(url, settings)
    }

    /**
     *
     * @param {Object} settings user export settings.
     * @param {string} chart chartConfig
     * @param {ClientRect} dimension bounding box of the main container.
     */

    function clientExport(settings, chart, dimension) {
        var cleanSVG = serializeChartSvg(chart);

        var exportFile = function() {
            switch (settings.type) {
                case "svg": {
                    handleSvgExport(settings, cleanSVG, dimension);
                    break
                }
                case "png":
                case "jpeg":
                case "bmp": {
                    // https://stackoverflow.com/questions/37412506/how-to-get-dataurl-from-blob-or-bloburl
                    handleImageExport(settings, cleanSVG, dimension);
                    break
                }
                case "pdf": {

                    var jsPDF = win.jsPDF;
                    if (defined(jsPDF) && typeof jsPDF === "function" && defined(jsPDF.API) && typeof jsPDF.API.addImage === "function") {

                        handleImageExport(settings, cleanSVG, dimension, function(imageData, scale) {
                            var margin = 0,
                                width = cleanSVG.width.baseVal.value + 2 * margin,
                                height = cleanSVG.height.baseVal.value + 2 * margin;

                            setTimeout(function() {
                                var pdf = new win.jsPDF('l', 'pt', [width, height]);
                                pdf.addImage(imageData, 'PNG', 0, 0, dimension.width, dimension.height);
                                pdf.save('chart.pdf');
                            }, 100)
                        });
                    } else {
                        Console.warn("Export type " + settings.type + " is not supported without jsPDF library");
                    }
                    break;
                }
                default: {
                    Console.warn("Export type " + settings.type + " is not supported");
                }
            }
            removeElement(cleanSVG);
        };


        if(!exporting) {
            exportFile();
        } else {
            var interval = setInterval(function() {
                if(!exporting) {
                    clearInterval(interval);
                    exportFile();
                }
            }, 1000/ 60)
        }
    }




    /**
     *
     * @param {Object} settings user export settings.
     * @param {string} chart chartConfig
     * @param {ClientRect} dimension bounding box of the main container.
     */

    function serverExport(settings, chart, dimension) {
        return post(settings.url, {
            "FILENAME": settings.filename,
            "TYPE": settings.type,
            "WIDTH": dimension.width * settings.scale,
            "HEIGHT": dimension.height * settings.scale,
            "DATA": JSON.stringify(chart.systemConf)
        }, settings.formAttributes);
    }

    var each = function (arr, fn) {
        return Array.prototype.forEach.call(arr, fn);
    };

    /*
     * Recursively merge properties of two objects
     */

    function merge() {
        var i,
            len = arguments.length,
            ret = {},
            doCopy = function (copy, original) {
                var key;
                // An object is replacing a primitive
                if (typeof copy !== 'object') {
                    copy = {};
                }
                for (key in original) {
                    if (original.hasOwnProperty(key)) {
                        copy[key] = original[key];
                    }
                }
                return copy;
            };

        for (i = 0; i < len; i++) {
            ret = doCopy(ret, arguments[i]);
        }
        return ret
    }

    /**
     * @param obj
     * @returns {boolean}
     */

    function defined(obj) {
        return obj !== undefined && obj !== null;
    }

    /**
     * Extend an object with the members of another
     * @param {Object} a The object to be extended
     * @param {Object} b The object to add to the first one
     */

    var extend = function (a, b) {
        var attr;
        if (!a) {
            a = {};
        }
        for (attr in b) {
            a[attr] = b[attr];
        }
        return a;
    };

    /**
     * @param {String} tag
     * @param {Object} attributes
     * @param {Object} styles
     * @param {Element} parent
     * @returns {Element}
     */

    function createElement(tag, attributes, styles, parent) {
        var el = document.createElement(tag);
        if (attributes) {
            extend(el, attributes);
        }
        if (styles) {
            extend(el.style, styles);
        }
        if (parent) {
            parent.appendChild(el);
        }
        return el;
    }

    /**
     * Removes the element.
     * @param {Element} element
     */

    function removeElement(element) {
        // create a garbage bin element, not part of the DOM
        if (!garbageBin) {
            garbageBin = createElement('div');
        }
        // move the node and empty bin
        if (element) {
            garbageBin.appendChild(element);
        }
        garbageBin.innerHTML = '';
    }


    /**
     *
     * @param {String} url
     * @param {Object} data
     * @param {Object} formAttributes
     */

    var post = function (url, data, formAttributes) {
        var name, form;
        // create the form
        form = createElement('form', merge({
            method: 'post',
            action: url
        }, formAttributes), {
            display: 'none'
        }, document.body);

        // add the data
        for (name in data) {
            createElement('input', {
                type: 'hidden',
                name: name,
                value: data[name]
            }, null, form);
        }

        // submit
        form.submit();

        // clean up
        removeElement(form);
    };

    global.exportChart = exportChart;

})(window, $ZC);