(function (ZC, d3, win) {
    
    /**
     * This file is part of Zoho Charts Library.
     */
    
    "use strict";

    var each = function (arr, fn) {
        return Array.prototype.forEach.call(arr, fn);
    };

    /**
     * Check for array
     * @param {Object} obj
     */
    function isArray(obj) {
        return Object.prototype.toString.call(obj) === '[object Array]';
    }

    /**
     *
     * @param obj
     * @return {boolean}
     */

    function defined(obj) {
        return obj !== undefined && obj !== null;
    }

    /**
     *
     * @param {String} str
     * @param {boolean} inside
     * @return {String}
     */

    var trim = function (str, inside) {
        if (typeof str === 'string') {
            str = str.replace(/^\s+|\s+$/g, '');
            if (inside && /^[0-9\s]+$/.test(str)) {
                str = str.replace(/\s/g, '');
            }
        }
        return str;
    };

    // for cross browser xml parsing support.
    var parseDom = (function () {
        if (typeof win.DOMParser != "undefined") {
            return function (xmlStr) {
                return ( new win.DOMParser() ).parseFromString(xmlStr, "text/xml");
            };
        } else if (typeof win.ActiveXObject != "undefined" &&
            new win.ActiveXObject("Microsoft.XMLDOM")) {
            return function (xmlStr) {
                var xmlDoc = new win.ActiveXObject("Microsoft.XMLDOM");
                xmlDoc.async = "false";
                xmlDoc.loadXML(xmlStr);
                return xmlDoc;
            };
        } else {
            throw new Error("No XML parser found");
        }
    })();

    /**
     *
     * @param {String} val
     * @return {Number|String|null}
     */

    var parseData = function (val) {
        var trimVal = trim(val),
            trimInsideVal = trim(val, true),
            floatVal = parseFloat(trimInsideVal);

        if (+trimInsideVal === floatVal) { // numeric
            return floatVal;
        } else {
            return trimVal === '' ? null : trimVal;
        }
    };

    /**
     *
     * @param {Array} rows column
     * @param {String} title
     * @param {Array} xAxis column
     * @param {String} chartType
     * @return {{type: (*|string), seriesname: (*|string), data: Array}}
     */

    function columnToZCSeries(rows, title, xAxis, chartType) {
        var series = {
            type: chartType,
            seriesname: title || "",
            data: []
        };

        each(rows, function (point, i) {
            series.data[i] = [xAxis[i], point]
        });
        return series;
    }

    /**
     *
     * @param sourceInfo user source info
     * @return {String|null}
     */

    function getSourceType(sourceInfo) {
        var keys = Object.keys(sourceInfo);
        var optionsIndex = keys.indexOf("options");

        // remove options
        if (optionsIndex > -1) {
            keys.splice(optionsIndex, 1);
        }
        return keys.pop();
    }


    /**
     *
     * If the source is defined as an array and it exists in the row returns that or if it is a string returns  same for all and
     * if it is not defined returns default as 'bar'.;
     * @param {Array|String|null} source
     * @param {Number} row
     * @return {string}
     */
    function getChartType(source, row) {
        var type;
        if (isArray(source)) {
            type = source[row];
        } else if (typeof source === "string") {
            type = source;
        }
        return type || "bar";
    }


    /**
     *
     * @param {Object} details
     * @param {String} type (csv|tsv|table|xml)
     * @return {Parser}
     * @constructor
     */

    var Parser = function (details, type) {
        if (!(this instanceof Parser)) {
            return new Parser(details, type);
        }
        this.init(details, type.toLowerCase());
    };


    Parser.prototype = {
        init: function (details, type) {

            this.source = details[type];
            this.options = details.options || {};
            this.columns = [];
            this.firstRowAsNames = defined(this.options.firstRowAsNames) || true;
            this.xAxisIndex = this.options.xAxisIndex || 0;
            switch (type) {
                case "table": {
                    this.parseTable();
                    break;
                }
                case "tsv":
                case "csv": {
                    this.parseCSV();
                    break;
                }
                case "xml": {
                    this.parseXML();
                    break;
                }
                default: {
                    this.error = true;
                }
            }
        },

        parseTable: function () {
            var options = this.options,
                table = this.source,
                columns = this.columns,
                startRow = options.startRow || 0,
                endRow = options.endRow || Number.MAX_VALUE,
                startColumn = options.startColumn || 0,
                endColumn = options.endColumn || Number.MAX_VALUE;

            if (table) {
                if (typeof table === "string") {
                    table = document.getElementById(table);
                }

                d3.select(table).selectAll("tr").each(function (tr, rowNo) {
                    if (rowNo >= startRow && rowNo <= endRow) {
                        each(this.cells, function (item, colNo) {
                            if ((item.tagName === 'TD' || item.tagName === 'TH') && colNo >= startColumn && colNo <= endColumn) {
                                if (!columns[colNo - startColumn]) {
                                    columns[colNo - startColumn] = [];
                                }
                                columns[colNo - startColumn][rowNo - startRow] = item.textContent;
                            }
                        })
                    }
                });
                if (columns.length) {
                    this.parseTypes();
                }
            }
        },

        parseCSV: function () {
            var csv = this.source,
                options = this.options,
                columns = this.columns || [],
                startRow = options.startRow || 0,
                endRow = options.endRow || Number.MAX_VALUE,
                startColumn = options.startColumn || 0,
                endColumn = options.endColumn || Number.MAX_VALUE,
                itemDelimiter,
                activeRowNo = 0;

            if (csv) {
                var lines = csv
                    .replace(/\r\n/g, '\n') // Unix
                    .replace(/\r/g, '\n') // Mac
                    .split(options.lineDelimiter || '\n');
                itemDelimiter = options.itemDelimiter || (csv.indexOf('\t') !== -1 ? '\t' : ',');

                each(lines, function (line, rowNo) {
                    var trimmed = trim(line),
                        isComment = trimmed.indexOf('#') === 0,
                        isBlank = trimmed === '',
                        items;

                    if (rowNo >= startRow && rowNo <= endRow && !isComment && !isBlank) {
                        items = line.split(itemDelimiter);
                        each(items, function (item, colNo) {
                            if (colNo >= startColumn && colNo <= endColumn) {
                                if (!columns[colNo - startColumn]) {
                                    columns[colNo - startColumn] = [];
                                }
                                columns[colNo - startColumn][activeRowNo] = item;
                            }
                        });
                        activeRowNo += 1;
                    }
                });
                if (columns.length) {
                    this.parseTypes();
                }
            }
        },

        parseXML: function () {

            // xml is directly converted to ZOHO Charts type.
            var self = this,
                xml = self.source,
                options = self.options,
                seriesdata = [],
                itemDelimiter = options.itemDelimiter || ',',
                chartTypes = options.chartTypes,
                row = 0;

            if (typeof xml === "string") {
                xml = parseDom(xml);
            }

            var serialize = function (node) {
                var dataNodes = node.selectAll("data"),
                    point, series;

                if (!dataNodes.empty()) {
                    series = {
                        seriesname: node.attr("name"),
                        data: [],
                        type: node.attr("type") || getChartType(chartTypes, row++),
                        disabled: (node.attr("disabled") === "true"),
                        yaxiscolumnorder: node.attr("yaxiscolumnorder") && node.attr("yaxiscolumnorder").split(",").map(Number)
                    };
                }

                dataNodes.each(function (d, i) {
                    point = series.data[i] = [];
                    var text = d3.select(this).text();
                    text.split(itemDelimiter).forEach(function (item, j) {
                        point[j] = parseData(item);
                    })
                });
                return series;
            };

            d3.select(xml).selectAll("series").each(function (e, j) {
                var series = serialize(d3.select(this));
                if (defined(series)) {
                    seriesdata[j] = series;
                }
            });
            if (seriesdata.length) {
                this.seriesdata = seriesdata;
            }
        },

        parseTypes: function () {
            var columns = this.columns,
                col = columns.length;
            if (col) {
                while (col--) {
                    this.parseColumn(columns[col]);
                }
            }
        },

        parseColumn: function (column) {
            var row = column.length;
            while (row--) {
                column[row] = parseData(column[row]);
            }
        },

        parse: function () {
            if (this.seriesdata) {
                // for xml return the data
                return this.seriesdata;
            }

            if (this.error) {
                return null;
            }

            var xAxisIndex = this.xAxisIndex,
                columns = this.columns,
                firstRowAsNames = this.firstRowAsNames,
                xColumn = columns.splice(xAxisIndex, 1).shift(),
                col = columns.length,
                seriesTitle,
                currentCol,
                seriesdata = [],
                currentChartType,
                chartTypes = this.options.chartTypes;

            if (firstRowAsNames) {
                xColumn.shift(); // remove the top left corner.
            }

            while (col--) {
                currentCol = columns[col];
                currentChartType = getChartType(chartTypes, col);
                if (firstRowAsNames) {
                    seriesTitle = currentCol.shift();
                }
                seriesdata[col] = columnToZCSeries(currentCol, seriesTitle, xColumn, currentChartType);
            }
            return seriesdata;
        }
    };

    /**
     *
     * @param {Object} sourceInfo
     * @returns {Array}
     */

    var parse = ZC.parseExternalData = function (sourceInfo) {
        var type = getSourceType(sourceInfo);
        if (type) {
            return new Parser(sourceInfo, type).parse();
        }
        return null;
    };
    
    ZC.chartFromExternalData = function (sourceInfo, chartConfig) {
        var series = parse(sourceInfo), container = sourceInfo.container;
        if(!defined(chartConfig.seriesdata)) {
            chartConfig.seriesdata = {};
        }
        chartConfig.seriesdata.chartdata = series;
        
        if(typeof container === "string") {
            container = d3.select("#" + container).node();
        }
        
        return new $ZC.charts(container, chartConfig);
    };

})($ZC, d3, window);