(function (global) {

    "use strict";

    /**
     * A draggable plugin for ZOHOCHARTS
     */

    var d3 = window.d3, SVGElement = window.SVGElement;

    var dragged = false;

    function defined(obj) {
        return obj !== undefined && obj !== null;
    }

    /**
     *
     * @param {Element|*} self
     * @returns {boolean}
     */

    function isLegend(self) {
        return /legendarea/.test(self.className);
    }

    function invoke(callback, dragObj) {
        if (typeof callback === "function") {
            callback.call({
                "x": dragObj.x,
                "y": dragObj.y
            }, dragObj.chart);
        }
    }

    function dragPossible(self) {
        var data = self.__data__ || {};
        return defined(data.x) && defined(data.y);
    }

    function dragInit() {
        if (!dragPossible(this)) {
            return;
        }

        var reference = this.__dragObj__.chart.svg.node().parentNode;
        var referenceCoordinates = reference.getBoundingClientRect();
        var elementCoordinates = this.getBoundingClientRect();
        global.utils.mergeJSON(this.__dragObj__, {
            "referenceWidth": referenceCoordinates.width,
            "referenceHeight": referenceCoordinates.height,
            "x": elementCoordinates.left - referenceCoordinates.left,
            "y": elementCoordinates.top - referenceCoordinates.top,
            "width": elementCoordinates.width,
            "height": elementCoordinates.height
        }, true, true);

        var dragObj = this.__dragObj__,
            chart = dragObj.chart,
            eventElement = dragObj.eventElement,
            callback = dragObj.events.start;
        this.style.cursor = "move";

        if (chart.isAxisCategory || chart.dataObject.isPolarAxisCategory) {
            eventElement.style.pointerEvents = "none";
        } else {

            chart.svg.select(".seriesholder").style({
                "pointer-events": "none"
            })
        }
        invoke(callback, dragObj);
    }

    function dragMove() {
        if (!dragPossible(this)) {
            return;
        }

        var element = d3.select(this);
        var d = this.__dragObj__, data = this.__data__, callback = d.events.move, userLegend = d.chart.userdata.legend;

        if (d.isSVG) {
            element
                .attr("x", data.x = d.x = Math.max(0, Math.min(d.referenceWidth - d.width, d3.event.x)))
                .attr("y", data.y = d.y = Math.max(0, Math.min(d.referenceHeight - d.height, d3.event.y)));
        } else {
            element.style({
                "left": function () {
                    data.x = d.x = Math.max(0, Math.min(d.referenceWidth - d.width, d3.event.x));
                    return d.x + "px";
                },
                "top": function () {
                    data.y = d.y = Math.max(0, Math.min(d.referenceHeight - d.height, d3.event.y));
                    return d.y + "px";
                }
            });
        }

        if (isLegend(this)) {
            this.firstElementChild.style.pointerEvents = "none";
            userLegend.x = data.x;
            userLegend.y = data.y;
        }

        invoke(callback, d);
        dragged = true;
    }

    function dragEnd() {
        var dragObj = this.__dragObj__,
            chart = dragObj.chart,
            eventElement = dragObj.eventElement;

        this.style.cursor = "auto";

        if (chart.isAxisCategory || chart.dataObject.isPolarAxisCategory) {
            eventElement.style.pointerEvents = "visible";
        } else {
            dragObj.chart.svg.select(".seriesholder").style({
                "pointer-events": "visible"
            })
        }

        if (!dragged) {
            return;
        }

        if (isLegend(this)) {
            this.firstElementChild.style.pointerEvents = "auto";
        }

        if (dragPossible(this) && dragged) {
            invoke(dragObj.events.end, dragObj);
        }
        dragged = false;
    }

    // Drag behavior

    var drag = d3.behavior.drag()
        .origin(function () {
            var temp, x, y;
            if (this.__dragObj__.isSVG) {
                temp = this.getBBox();
                x = temp.x;
                y = temp.y;
            } else {
                x = this.offsetLeft;
                y = this.offsetTop;
            }
            return {x: x, y: y};
        })
        .on("dragstart", dragInit)
        .on("drag", dragMove)
        .on("dragend", dragEnd);


    function draggable(element, chartObj, callbacks) {
        if (element.node().__dragObj__ !== undefined) {
            return false;
        }
        var tmp_elm = element.node();
        // set drag properties
        tmp_elm.__dragObj__ = {
            "isSVG": tmp_elm instanceof SVGElement,
            "eventElement": chartObj.eventHandler.trackerElement.node(),
            "events": callbacks || {},
            "chart": chartObj
        };
        element.call(drag);
        return true
    }

    Object.defineProperty(global, "draggable", {
        "value": draggable
    });
    // global.draggable = draggable;
})($ZC);
