;

(function(ZC) {
    "use strict";

    var axis = ZC.axis;

    /**
     *
     * @param {Object} obj
     * @returns {{top, right, bottom, left}}
     */

    function getMargin(obj) {
        return {
            top: pick(obj.marginTop, 5),
            right: pick(obj.marginRight, 5),
            bottom: pick(obj.marginBottom, 5),
            left: pick(obj.marginLeft, 5)
        }
    }


    /**
     *
     * @param {Number} rotation in degree's
     * @param {Number} side in TRBL direction as 0 to 3
     * @returns {String}
     */

    function autoLabelAlign(rotation, side) {
        var ret,
            angle = (pick(rotation, 0) - (side * 90) + 720) % 360;
        if (angle > 15 && angle < 165) {
            ret = 'end';
        } else if (angle > 195 && angle < 345) {
            ret = 'start';
        } else {
            ret = 'middle';
        }
        return ret;
    }

    /**
     *
     * @param {Array} categories
     * @param {Number} expectedCount
     * @param {Number} interval
     * @param {Boolean} filterByIndices
     * @returns {Array}
     */

    function filterCategories(categories, expectedCount, interval, filterIndex) {
        var actualCount = categories.length;
        expectedCount = mathRound(expectedCount);
        interval = interval || mathCeil(actualCount / expectedCount);
        var temp = 0;
        var tickvalues = [], tickvalueIndices = (filterIndex) ? [] : null, tickvalueNodes = (filterIndex) ? [] : null;
        while (temp < actualCount) {
            if(filterIndex){
                tickvalues.push(categories[temp].value);
                tickvalueIndices.push(categories[temp].index);
                tickvalueNodes.push(categories[temp]);
            }else{
                tickvalues.push(categories[temp]);
            }
            temp = temp + interval
        }

        if(filterIndex){
            return {
                "nodes" : tickvalueNodes,
                "ticks" : tickvalues,
                "indices" : tickvalueIndices
            };
        }else{
            return tickvalues;
        }
    }

    /**
     *
     * @param formattedTickValues
     * @param tickConfig
     * @param columnInfo
     * @param {Chart} chart
     * @param {Number} columnIndex
     * @param {Number} levelIndex
     * @returns {{tickValues: Array, styles: Array, userRawTicks: Array}}
     */
    function getCustomTickValues(formattedTickValues, tickConfig, columnInfo, chart, columnIndex, levelIndex) {

        var parentArray = [], parentStyles = [], userTickValues = [];

        var userColumnInfo = {
            data: [columnInfo]
        };
        columnInfo.columnindex = pick(columnInfo.columnindex, columnIndex);

        formattedTickValues.forEach(function(d, i) {
            var userInput = invokeFunction(tickConfig.showAs, d, tickConfig, userColumnInfo, chart, i, levelIndex),
                childArray = [], styles = [];

            if (isArray(userInput)) {
                parentArray.push(childArray);
                parentStyles.push(styles);
            } else {
                childArray = parentArray;
                styles = parentStyles
            }
            userTickValues.push(userInput);

            if (defined(userInput)) {
                var sandBox = document.createElement("div");
                sandBox.innerHTML = userInput;

                var tspan = toArray(sandBox.querySelectorAll("tspan")),
                    images = sandBox.querySelectorAll("image,img");

                if(tspan.length) {
                    tspan.forEach(function(t) {
                        childArray.push(t.textContent.trim());
                        styles.push({
                            fontSize: t.style.fontSize && parseFloat(t.style.fontSize),
                            fontFamily: t.style.fontFamily,
                            fontStyle: t.style.fontStyle,
                            fontWeight:t.style.fontWeight
                        })
                    });
                } else {
                    childArray.push(sandBox.textContent);
                }

                if(images.length) {
                    toArray(sandBox.querySelectorAll("image,img")).forEach(function(t) {
                        childArray.push({
                            width: +t.getAttribute("width"),
                            height: +t.getAttribute("height")
                        });
                        styles.push({})
                    });
                }
            }
        });

        return {
            tickValues: parentArray,
            styles: parentStyles,
            userRawTicks: userTickValues
        }
    }


    /**
     *
     * @param {Array} ticks formatted tick values
     * @param {Array} ticksOriginal raw tick values
     * @param {Number} fontSize font-size of the tick label
     * @param {String} fontFamily font-family of the tick label
     * @param {{fontFamily: string, fontSize: number, fontStyle: string}} styles
     * @param {Boolean} calculateAsArea
     * @returns {{actualVal: *, formattedVal: String}}
     */

    function getLongestTick(ticks, ticksOriginal, fontSize, fontFamily, styles, calculateAsArea) {

        /**
         * @type {CanvasRenderingContext2D} tempCanvasContext
         */
        tempCanvasContext.save();
        tempCanvasContext.font = fontSize + "px " + fontFamily;

        var temp = 0,
            longest = null,
            original = null,
            style,
            index;

        styles = pick(styles, []);

        if (!defined(ticksOriginal)) {
            ticksOriginal = [];
        }

        var check = function(label, i) {
            var bound = getTextBound(label, fontSize, fontFamily, styles[i]),
                width = bound.width;

            if(calculateAsArea) {
                width = bound.width * bound.height;
            }

            if (width > temp) {
                temp = width;
                longest = label;
                original = ticksOriginal[i];
                style = styles[i];
                index = i;
            }
        };
        ticks.forEach(check);
        tempCanvasContext.restore();
        return {
            actualVal: original,
            formattedVal: longest,
            style: style,
            index: index
        };
    }


    /**
     * returns a transform string
     *
     * @param {Number} rotate in degree's
     * @param {Number} x
     * @param {Number} y
     * @returns {string}
     */
    function generateTransform(rotate, x, y) {
        return "rotate(" + rotate + " " + (x || 0) + " " + (y || 0) + ")";
    }

    /**
     * returns the size of the tick label in horizontally or vertically
     *
     * @param {String} side "width" | "height"
     * @param {Object} dimension
     * @param {Object} margin
     * @param {Number} rotation in degree's
     * @param tickFontSize
     * @param orient
     * @returns {number}
     */

    function getTickWidth(side, dimension, margin, rotation, tickFontSize, orient) {
        var space = 0;
        if (side === "height") {
            space = Math.abs(tickFontSize + margin.top + margin.bottom);
        } else {
            if (rotation > 0) {
                space = distBetweenRotatedRect(rotation, dimension.width, tickFontSize, orient).tickSpacing + margin.left + margin.right
            } else {
                space = dimension.width + margin.left + margin.right;
            }
        }
        return space;
    }

    /**
     * returns the opposite side margin based on orien
     * @param {String} orient
     * @param {Object} margin
     */

    function getOppositeMargin(orient, margin) {
        return margin[getOppositeOrient(orient)];
    }

    /**
     *
     * @param {String} orient
     * @param {Object} tickMargin
     * @param {Object} tickMarkConfig
     * @returns {{x: number, y: number}}
     */

    function getTickMargin(orient, tickMargin, tickMarkConfig) {


        var tickMarkSize = pick(tickMarkConfig.size, 6),
            margin = getOppositeMargin(orient, tickMargin),
            prop = {
                x: 0, y: 0
            };

        switch (orient) {
            case "left":
                prop.x = -(margin + tickMarkSize);
                break;
            case "top":
                prop.y = -(margin + tickMarkSize);
                break;
            case "right":
                prop.x = (margin + tickMarkSize);
                break;
            default:
                prop.y = (margin + tickMarkSize);
                break;
        }
        return prop;
    }

    /**
     * returns a empty dimension.
     * @returns {{}}
     */
    function getEmptyBBox() {
        var t = {};
        t.x = t.y = t.width = t.height = 0;
        return t;
    }

    function distBetweenRotatedRect(rotation, width, height, orient) {
        var ang1 = rotation * (mathPI * 2 / 360);
        var ang2 = (90 - rotation) * (mathPI * 2 / 360);

        if (orient === "bottom" || orient === "top") {
            return {
                "tickSpacing": mathAbs(height / mathSin(ang1)) + 5,
                "baseLen": 0
            }
        } else {
            var len1 = mathAbs(mathSin(ang2) * height);
            var len2 = mathAbs((mathCos(ang2) * height) * mathTan(ang1));
            return {
                "tickSpacing": len1 + len2 + 5,
                "baseLen": mathAbs(mathCos(ang2) * width)
            }
        }
    }

    function addMinmaxToTickValues(tickvalues, axisticklabeldata, chartObj, type, axisorder, orient) {
        var dataObject = chartObj.dataObject;
        var minmaxlabelsdata = axisticklabeldata.minmaxlabels || {};
        if (minmaxlabelsdata.show === true) {
            var newScale = $ZC.scale.getScale(chartObj, dataObject, chartObj.plotarea, orient, axisorder, chartObj.dataset, type); //Scale with actual outerPadding
            var domain = newScale.domain();
            if (validateNullVal(minmaxlabelsdata.excludeIntervals, true)) {
                tickvalues = domain;
            } else {
                if (domain[0] !== tickvalues[0]) {
                    tickvalues.splice(0, 0, domain[0]);
                }
                if (domain[1] !== tickvalues[tickvalues.length - 1]) {
                    tickvalues.push(domain[1]);
                }
            }
        }
        return tickvalues;
    }

    function nearestNiceMin(minDomain, remaining, magnitude, interval) {
        var arr = [1, 2, 5, 10], len = arr.length;
        var actualMin = minDomain + (remaining / 2);
        actualMin = +actualMin.toFixed(12);
        var diff, nearestMin;
        for (var j = 0; j < len; j++) {
            var a1 = magnitude / arr[j], curMin;
            if (minDomain % a1) {
                curMin = minDomain - (minDomain % a1) + a1;
            } else {
                curMin = minDomain;
            }

            var curDiff = mathAbs(actualMin - curMin);
            if (curDiff < diff || diff == null) {
                diff = curDiff;
                nearestMin = curMin;
            }
        }
        if (nearestMin == null || nearestMin < minDomain) {
            var e = mathFloor(mathLog(interval) / mathLN10);
            if (e < 0) {
                nearestMin = +minDomain.toFixed(mathAbs(mathMin(0, e)));
            }
            nearestMin = minDomain;
        }
        nearestMin = (nearestMin > minDomain) ? nearestMin : minDomain;
        return nearestMin;
    }

    function nearestTickInterval(actualInterval, magnitude, tickCount) {
        var curDiff, prevDiff = null, curInterval, nearestInterval = magnitude;
        var multiples = [1, 1.2, 1.5, 1.75, 2, 2.2, 2.5, 3, 3.5, 4, 4.5, 5, 6, 7, 8, 9, 10];
        for (var i = 0; i < multiples.length; i++) {
            curInterval = multiples[i] * magnitude;
            curDiff = mathAbs(actualInterval - curInterval);
            prevDiff = prevDiff == null ? curDiff : prevDiff;
            if (curDiff < prevDiff && curInterval <= actualInterval) {
                prevDiff = curDiff,
                    nearestInterval = curInterval;
            }
        }
        return nearestInterval;
    }


    function getNiceMin(domain, minStep) {
        var step = getLinearStep(domain);
        if (defined(minStep)) { // Test_525.json
            step = mathMax(minStep, step);
        }
        return mathCeil(domain[0] / step) * step;
    }

    function getNiceStep(domain) {
        return mathCeil(getLinearStep(domain));
    }

    function getLinearStep(domain, count) {
        var m = defined(count) ? count : 10,
            span = domain[1] - domain[0],
            step = mathPow(10, mathFloor(mathLog(span / m) / mathLN10)),
            err = m / span * step;
        if (err <= .15) {
            step *= 10;
        }
        else if (err <= .35) {
            step *= 5;
        }
        else if (err <= .75) {
            step *= 2;
        }
        return step;
    }


    /**
     * Axis Label Rendering
     *
     * @param {Object} axisConfig
     * @param {Object} otherInfo
     * @param {String} orient
     * @param {Object} parent
     * @constructor
     */


    var AxisLabel = function(axisConfig, otherInfo, orient, parent) {
        this.axis = axisConfig;
        this.config = axisConfig.label;
        this.axisRect = otherInfo.bound;
        this.parentNode = parent.isD3Selection ? parent : d3.select(parent);
        this.orient = orient;
        this.systemConfig = otherInfo.systemConf;
        this.order = otherInfo.axisOrder;
        this.plotRect = otherInfo.plotRect;
        this.show = !otherInfo.hide;
        if (this.show) {
            this.add();
        } else {
            this.element = appendEle(this.parentNode, "text", [], ".", "class", "axislabel"); // a empty selection
        }

    };

    AxisLabel.prototype = {

        add: function() {
            var label = this;
            label.element = appendEle(label.parentNode, "text", [1], ".", "class", "axislabel");
            label.text = label.config.text;
            label.element.text(label.text);
            label.rect = null;
            label.transformLabel();
        },

        remove: function() {
            var label = this;
            label.rect = null;
            label.element = appendEle(label.parentNode, "text", [], ".", "class", "axislabel"); // remove it
        },

        setStyles: function() {
            var label = this,
                systemConfig = label.systemConfig;
            applyFont(label.element, [label.config, systemConfig.chart, systemConfig.canvas, $ZCG]);
        },

        getBBox: function() {
            var label = this;
            if (!defined(label.rect)) {
                if (label.element.isD3Selection && !label.element.empty()) {
                    label.rect = label.element.node().getBBox(); // calculate and cache.
                } else {
                    return getEmptyBBox();
                }
            }
            return label.rect;
        },

        transformLabel: function(axisBound, plotBound, side) {

            var label = this;

            if (label.show !== true) {
                return label.remove();
            }

            if ((label.element.empty() && label.show === true) || (label.text !== label.config.text)) {
                return label.add();
            }

            label.setStyles();

            var rotations = {
                    top: 0,
                    right: 90,
                    bottom: 0,
                    left: 270
                },
                systemConfig = label.systemConfig,
                textAnchor = "middle", // it is middle for all angle
                orient = pick(side, label.orient),
                rotate = rotations[orient],
                x,
                y,
                axisRect = pick(axisBound, this.axisRect),
                plotRect = pick(plotBound, this.plotRect),
                tickLabelConfig = label.axis.ticklabel,
                axisLabelConfig = label.config,
                tickMargin = getMargin(tickLabelConfig),
                labelMargin = getMargin(axisLabelConfig),
                fontSize = parseFloat($ZC.utils.getVal(getValByPriority([label.config, systemConfig.chart, systemConfig.canvas, $ZCG], "fontSize")));

            // reset BBox
            if(label.fontSize !== fontSize) {
                label.rect = null;
            } else {
                label.fontSize = fontSize;
            }

            // if axis rotated is called;
            label.element.attr("dy", "-0.25em");

            switch (orient) {
                case "top": {
                    x = plotRect.width / 2;
                    y = axisRect.height + tickMargin.top + labelMargin.bottom;
                    applyTransform(rotate, x, -y);
                    break;
                }

                case "right": {
                    // label.element.attr("dy", "-0.25em");
                    x = plotRect.height / 2;
                    y = axisRect.width + tickMargin.right + labelMargin.left;
                    applyTransform(rotate, x, -y);
                    break;
                }
                case "bottom": {
                    x = plotRect.width / 2;
                    y = axisRect.height + tickMargin.bottom+ labelMargin.top;

                    // apply baseline
                    label.element.attr("dy", "0.91em");
                    applyTransform(rotate, x, y);
                    break;
                }
                case "left": {
                    // label.element.attr("dy", "-0.25em");
                    x = axisRect.width + tickMargin.left + labelMargin.right;
                    y = plotRect.height / 2;
                    applyTransform(rotate, -x, y, -x, y);
                    break;
                }
            }

            function applyTransform(rotate, x, y, xTransform, yTransform) {
                xTransform = pick(xTransform, 0);
                yTransform = pick(yTransform, 0);

                label.element.attr({
                    "x": x,
                    "y": y,
                    "transform": generateTransform(rotate, xTransform, yTransform),
                    "text-anchor": textAnchor
                })
            }
        }
    };


    /**
     *
     * @param {String} type x or y
     * @param axisInfo
     * @param plotRect
     * @param helpers
     * @param tickCount
     * @constructor
     */

    var Ticks = function(type, axisInfo, plotRect, helpers, tickCount) {

        this.type = type;
        this.config = axisInfo.ticklabel;
        this.axis = axisInfo;
        this.plotRect = plotRect;
        this.helpers = helpers;

        this.tickCount = tickCount; // for web chart

        // max size
        this.maxSize = getMaxTickSize(this.config, plotRect);
        this.cache = {};
    };

    Ticks.prototype = {


        getScale: function() {
            var tick = this,
                chart = tick.helpers.chart,
                helpers = tick.helpers;
            return $ZC.scale.getScale(chart, chart.dataObject, tick.plotRect, helpers.orient, helpers.axisOrder, chart.dataset, tick.type, tick.extent);
        },

        getTicks: function(plotArea, needCount, helper) {

            var tick = this,
                helpers = pick(helper, tick.helpers),
                chart = helpers.chart,
                systemConf = chart.systemConf,
                dataObject = chart.dataObject,
                plotRect = tick.plotRect = pick(plotArea, tick.plotRect),
                axis = tick.axis,
                type = tick.type,
                userTickInterval = axis.interval,
                userTickCount = axis.tickCount,
                range,
                domain,
                scale,
                tickInterval = axis.interval,
                tickCount = pick(needCount, 0),
                extent = helper.extent,
                tickValues = extent,
                categoriesList, filterByIndex, tickValueIndices,
                dontFilter,
                formattedTickValues,

                axisOrder = helpers.axisOrder,
                axisRotated = helpers.axisRotated,
                dataType = chart.dataset.getDataType(type, axisOrder),
                metaData = chart.systemConf.metadata,
                columnIndex = chart.dataset.getColumnIdx(type, axisOrder),
                columnInfo = metaData.columns[columnIndex],
                originalDataType = metaData.columns[columnIndex].datatype,
                isMergedAxis = (type == "y") ? (metaData.axes.y[axisOrder].length > 1) : false,
                formatConfig = pick(((isMergedAxis && type == "y") ? merged_axis_column_intersect(axisOrder, chart) : metaData.columns[columnIndex][originalDataType]), {}),
                formattingSubFunction = formatConfig.subfunction,
                formattingFunction = (type === "y" && chart.dataObject.isStackedPercentChart) ? getFormattedValueByPercent : getFormattedValue,

                margin = getMargin(tick.config),
                tickMargin = getTickMargin(helpers.orient, margin, tick.axis.tickmark),
                axisSize,

                tickFontSize = parseFloat($ZC.utils.getVal(getValByPriority([tick.config, systemConf.chart, systemConf.canvas, $ZCG], "fontSize"))),
                tickFontFamily = $ZC.utils.getVal(getValByPriority([tick.config, systemConf.chart, systemConf.canvas, $ZCG], "fontFamily")),

                longestTick,
                side = isRotatedChart(type, axisRotated) ? "height" : "width",
                isAutoRotation = isTickAutoRotate(chart,this.type,axisOrder,tick.config),
                 isHierarchy = chart.axes.x.count > 1 && type == "x" && axisOrder > 0,
                rotation = isAutoRotation || isHierarchy ? 0 : +tick.config.rotation,

                firstChartName = chart.chartTypes[0].name,
                outerPadding = chart.systemConf.chart.plot.plotoptions[firstChartName].outerPadding || 0,
                pad = chart.systemConf.chart.plot.plotoptions[firstChartName].padding || 0.2,
                longestFormattedString = null,
                plotSize = getPlotSize(helpers.chart,side,plotRect),
                padding = type === "x" ? (plotSize * outerPadding) * 2 : 0,

                dimension, space, customTickInfo = {},


                hasTimeGroupBy = isArray(formatConfig.groupby) && formatConfig.groupby.length > 0 || (formatConfig.groupby === "AUTO" && formattingSubFunction !== "YEAR");
                var specifierIndex = (hasTimeGroupBy) ? axisOrder : null;

            if (!isAutoRotation) {
                if (rotation < 0) {
                    rotation = 360 - mathAbs(rotation);
                } else if (rotation > 360) {
                    rotation = mathAbs(rotation) % 360;
                }
            }

            switch (dataType) {

                case "numeric": {

                    domain = defined(extent) ? extent : type === "x" ? dataObject.getMinMaxX() : dataObject.getMinMaxY(helpers.axisOrder);

                    // define axis size and range
                    if (side === "height") {
                        axisSize = plotRect.height;
                        tickCount = tickCount || mathRound(axisSize / 40);
                        range = [axisSize, 0];
                    } else {
                        axisSize = plotRect.width;
                        tickCount = tickCount || mathRound(axisSize / 100);
                        range = [0, axisSize];
                    }

                    // define tick interval
                    if (defined(userTickCount)) {
                        tickInterval = (domain[1] - domain[0]) / userTickCount;
                    }

                    var count, step, min, max, useTickInterVal;
                    useTickInterVal = defined(userTickInterval) && ((domain[1] - domain[0]) / userTickInterval) < axisSize;

                    // numeric scale
                    scale = d3_scaleLinear().domain(domain).range(range);

                    // get ticks
                    if (useTickInterVal) {

                        count = mathFloor((domain[1] - domain[0]) / tickInterval);

                        if (count) {
                            step = getLinearStep(domain, count);
                            min = mathCeil(domain[0] / step) * step;

                            if ((min - domain[0]) > tickInterval) {
                                var s = mathPow(10, mathFloor(mathLog(tickInterval) / mathLN10)),
                                    q = mathFloor((min - domain[0]) / s);
                                min -= (q * s);
                            }
                        } else {
                            var span = domain[1] - domain[0],
                                s = mathPow(10, mathFloor(mathLog(span) / mathLN10));
                            min = domain[0] - (domain[0] % s) + s;
                        }

                        tickValues = d3_range(min, domain[1], tickInterval);

                    } else if (+userTickCount > 0) {

                        if (userTickCount === 1) {
                            step = getLinearStep(domain, 1);
                            min = Math.ceil(domain[0] / step) * step;
                            tickValues = [min];
                        } else {
                            var m = userTickCount - 1, magnitude, min, min , remaining;
                            span = domain[1] - domain[0];
                            magnitude = mathPow(10, mathFloor(mathLog(span / m) / mathLN10));

                            magnitude = magnitude * Math.floor((span / m) / magnitude);
                            magnitude = +magnitude.toFixed(12); //To avoid floating point error (e.g):1.7000000000000002

                            var nicedInterval = nearestTickInterval(tickInterval, magnitude, m);
                            min = domain[0];
                            max = domain[1];
                            remaining = span - (nicedInterval * m);

                            // if (remaining < magnitude && Math.floor((span / m) / magnitude) > 1) {
                            //     nicedInterval -= magnitude;
                            //     remaining = span - (nicedInterval * m);
                            // }

                            var mag_remaining = mathPow(10, mathFloor(mathLog(remaining) / mathLN10) + 1);
                            min = nearestNiceMin(min, remaining, mag_remaining,nicedInterval);
                            min = +min.toFixed(12); //To avoid floating point error (e.g):1.7000000000000002
                            tickValues = d3_range(min, (min + (nicedInterval*userTickCount)), nicedInterval);

                        }

                    } else if (formattingSubFunction === "integer") {
                        tickValues = d3_range(getNiceMin(domain, 1), domain[1], getNiceStep(domain));

                    } else {
                        tickValues = scale.ticks(tickCount);
                    }

                    // add min max label if needed
                    tickValues = addMinmaxToTickValues(tickValues, tick.config, chart, type, helpers.axisOrder, helpers.orient);

                    tickCount = (this.config.zigzag) ? tickValues.length/2 : tickValues.length;
                    break;
                }
                case "ordinalNonLinear":
                case "ordinal": {


                    var visibleTickCount, size = tickFontSize,
                        hierarchical = dataObject.hierarchical, tickNodes;
                        filterByIndex = defined(hierarchical) && originalDataType == ORDINAL_DATATYPE && type == "x" && axisOrder == 0,
                        dontFilter = filterByIndex && (helpers.chart.systemConf.chart.width == "auto" || helpers.chart.systemConf.chart.height == "auto");


                    if (!defined(tickValues)) {
                        tickValues = dataObject.getCategories(type, helpers.axisOrder).slice(0);
                    }
                    if(filterByIndex){
                        tickValues = dataObject.hierarchical.leafNodes;
                    }

                    if(originalDataType === TIME_DATATYPE && !defined(formattingSubFunction)) {
                            // set sub function
                        var info = getAutoTimeSubFunction(tickValues[tickValues.length - 1] - tickValues[0]);
                        formattingSubFunction = formatConfig.subfunction = info.subFunction;
                    }

                    tickCount = tickValues.length;

                    dimension = {};
                    dimension.width = dimension.height = tickFontSize;
                    space = getTickWidth(side, dimension, margin, rotation, tickFontSize, helpers.orient);
                    visibleTickCount = mathRound((plotSize- padding) / space); // for initial filter.

                    if (visibleTickCount < tickCount && side === "width") {
                        if (rotation || tick.config.rotation > 0) {
                            space = distBetweenRotatedRect(45, tickFontSize, size, helpers.orient).tickSpacing + margin.left + margin.right;
                        }
                    }
                    if((!(type == "x" && axisOrder > 0)) && (!dontFilter)){
                        if (userTickCount < tickCount) {
                            tickValues = filterCategories(tickValues, userTickCount, null, filterByIndex);
                        } else if (defined(tickInterval)) {
                            tickValues = filterCategories(tickValues, tickCount, tickInterval, filterByIndex);
                        } else if (visibleTickCount < tickCount) {
                            tickValues = filterCategories(tickValues, visibleTickCount, null, filterByIndex);
                        }
                    }
                    if(filterByIndex){
                        categoriesList = tickValues.nodes || tickValues,
                        tickValueIndices = tickValues.indices,
                        tickValues = tickValues.ticks || dataObject.getCategories(type, helpers.axisOrder).slice(0);
                    }

                    tickCount = (this.config.zigzag) ? tickValues.length/2 : tickValues.length;

                    break;
                }

                case "time": {

                    var isAutoFormat = hasAutoFormat(formatConfig);

                    if (isAutoFormat) {

                        var timeScale = $ZC.scale.getScale(chart, dataObject, plotRect, helpers.orient, helpers.axisOrder, chart.dataset, type, null);
                        tickValues = timeScale.ticks().map(function(d) {
                            return timeScale.tickFormat()(d)
                        });

                    } else {

                        var start, end, dateInterval;
                        domain = defined(extent) ? extent : type === "x" ? dataObject.getMinMaxX(0) : dataObject.getMinMaxY(helpers.axisOrder, 0);

                        start = new Date(domain[0]);
                        end = new Date(domain[1]);

                        if (!defined(formattingSubFunction) || formattingSubFunction === "AUTO") {
                            // set sub function
                            var info = getAutoTimeSubFunction(end - start);
                            formattingSubFunction = formatConfig.subfunction = info.subFunction;
                        }

                        // get the date utils 1326
                        dateInterval = new date_utils(formatConfig);
                        dateInterval.useUTC = formatConfig.useUTC === true;

                        longestFormattedString = dateInterval.getMaxDate(formatConfig, specifierIndex); // save it
                        dimension = getTextBound(longestFormattedString, tickFontSize, tickFontFamily);
                        space = getTickWidth(side, dimension, margin, rotation, tickFontSize, helpers.orient);

                        tickCount = (plotSize - padding) / space;
                        // tickCount = (this.config.zigzag) ? tickCount/2 : tickCount;
                        // get the date range
                        var dateRange = dateInterval.range(start, end, mathCeil(tickCount), formattingSubFunction, tickInterval, chart);
                        // add min max tick
                        tickValues = addMinmaxToTickValues(dateRange.TICKS, tick.config, chart, type, helpers.axisOrder, helpers.orient);

                        tickCount = (this.config.zigzag) ? dateRange.TICKS.length/2 : dateRange.TICKS.length;

                        break;
                    }
                }
            }


            if(dataType === "ordinalNonLinear"){
                var formattedTickValues = isAutoFormat ? tickValues : formatArray(tickValues, formattingFunction, formatConfig, chart, originalDataType, "axis", specifierIndex);
                // get custom tick labels
                if (defined(tick.config.showAs)) {
                    customTickInfo = getCustomTickValues(formattedTickValues, tick.config, columnInfo, chart, columnIndex, axisOrder);
                    formattedTickValues = customTickInfo.tickValues;
                    customTickInfo.length = formattedTickValues.length;
                }
                var longestTick = getLongestTick(formattedTickValues, tickValues, tickFontSize, tickFontFamily, customTickInfo.styles);
                return {
                    rawTickValues: tickValues,
                    formatterTickValues: formattedTickValues,
                    longestTickValue: longestTick,
                    scale: tick.getScale(),
                    longestTickDimension: getTextBound(longestTick, tickFontSize, tickFontFamily),
                    margin: margin,
                    tickMargin: tickMargin,
                    rotation: rotation,
                    customTickInfo: customTickInfo
                }
            }

            // get a scale
            scale = tick.getScale();

            //
            var actualRange = scale.actualRange;
            var isSingleTick = actualRange && (actualRange[0] === actualRange[1]);
            if (isSingleTick) {
                tickValues = [scale.invert(actualRange[0])];
            }

            // format the values and get the longest tick value.

            formattedTickValues = isAutoFormat ? tickValues : formatArray(tickValues, formattingFunction, formatConfig, chart, originalDataType, "axis", specifierIndex);


            // get custom tick labels
            if (defined(tick.config.showAs)) {
                customTickInfo = getCustomTickValues(formattedTickValues, tick.config, columnInfo, chart, columnIndex);
                formattedTickValues = customTickInfo.tickValues;
                customTickInfo.length = formattedTickValues.length;
            }

            longestTick = getLongestTick(formattedTickValues, tickValues, tickFontSize, tickFontFamily, customTickInfo.styles);

            // uncomment the bellow condition to use the default longest date.
            if(!defined(longestFormattedString)) {
                longestFormattedString = longestTick.formattedVal;
            }


            // calculate number of tick can be fit.
            var tickBound = getTextBound(longestFormattedString, tickFontSize, tickFontFamily, longestTick.style, true);

            if(dontFilter){
                tick.recalculatePlotSize(helpers.chart, plotSize, tickBound, side, tickCount, margin, pad, outerPadding)
            }

            var maxTickCount = tick.getMaxTickCount(tickBound, rotation, margin, outerPadding, side, helpers.orient, type,helpers.chart);

            if (!isSingleTick && maxTickCount < tickCount) {

                if (isAutoRotation && side === "width") {
                    if(helpers.chart.axes.x.count > 1 && type == "x" && axisOrder == 0){
                        rotation = 90; //For hierarchy auto rotate 90
                    }else{
                        rotation = 45;
                    }

                    maxTickCount = tick.getMaxTickCount(tickBound, rotation, margin, outerPadding, side, helpers.orient,null,helpers.chart);
                }

                // time data need a re-calculation of tick values after rotation

                if(dataType === TIME_DATATYPE) {
                    dateRange = dateInterval.range(start, end, maxTickCount, formattingSubFunction, tickInterval, chart);
                    tickValues = addMinmaxToTickValues(dateRange.TICKS, tick.config, chart, type, helpers.axisOrder, helpers.orient);
                    formattedTickValues = formatArray(tickValues, formattingFunction, formatConfig, chart, originalDataType, "axis",specifierIndex);

                } else if (maxTickCount < tickCount) {
                    // filter ticks if needed
                    if(filterByIndex){
                        tickValues = filterCategories(categoriesList || tickValues, maxTickCount, null,filterByIndex);
                        tickValueIndices = tickValues.indices,
                        tickValues = tickValues.ticks;
                    } else{
                        tickValues = filterCategories(tickValues, maxTickCount);
                        if(tick.config.showAs) {
                            // TODO have convert multiple object to single object to reduce loop.
                            customTickInfo.styles = filterCategories(customTickInfo.styles, maxTickCount);
                            customTickInfo.userRawTicks = filterCategories(customTickInfo.userRawTicks, maxTickCount)
                        }
                    }

                    formattedTickValues = filterCategories(formattedTickValues, maxTickCount);
                    longestTick = getLongestTick(formattedTickValues, tickValues, tickFontSize, tickFontFamily, customTickInfo.styles);
                }
            }


            // time data group by
            if (originalDataType === TIME_DATATYPE && hasTimeGroupBy) {
                if (hasTimeGroupBy && !defined(dataObject.hierarchical)) {
                    dataObject.hierarchical = new $ZC.HierarchicalTimeDataAdapter(chart, tickValues, formatConfig.groupby, dataType !== TIME_DATATYPE, columnInfo);
                } else if(hasTimeGroupBy) {
                    dataObject.hierarchical.update(tickValues);
                }
            }


            if(defined(customTickInfo.styles)) {
                longestTick = getLongestTick(formattedTickValues, tickValues, tickFontSize, tickFontFamily, customTickInfo.styles, true);
            }

            var textDimension = getTextBound(longestFormattedString, tickFontSize, tickFontFamily, longestTick.style);

            return {
                rawTickValues: tickValues,
                formatterTickValues: formattedTickValues,
                longestTickValue: longestTick,
                scale: scale,
                longestTickDimension: textDimension,
                margin: margin,
                tickMargin: tickMargin,
                rotation: rotation,
                tickValueIndices : tickValueIndices,
                customTickInfo: customTickInfo
            }
            // now handle over flow of the formatted ticks and return;
        },


        getMaxTickCount: function(tickBound, rotation, margin, outerPadding, side, orient, type, chartObj) {
            var tick = this,
                tickArea,
                temp,
                tickCount,
                padding = 0,
                countForAngle;

            if (rotation === "auto") {
                rotation = 0;
            }

            countForAngle = !!rotation;

            var length = getPlotSize(chartObj,side,tick.plotRect)
            if (type === "x") {
                padding = (length * outerPadding) * 2;
            }

            tickArea = length - padding;


            if (countForAngle) {

                if (side === "height" && (rotation === 90 || rotation === 270)) {
                    return tickArea / (tickBound.width + mathAbs(margin.top) + mathAbs(margin.bottom));
                }

                // var angleCheck1 = side === "height" ? (rotation !== 90 && rotation !== 270 && rotation !== 0 && rotation !== 180 && rotation !== 360) :
                //     (rotation !== 0 && rotation !== 180 && rotation !== 360);

                var r = rotation % 90;
                var angleCheck = side === "height" ? (r > 0) : (r > 0 || rotation === 90 || rotation === 270);

                if (angleCheck) {
                    var distBwTicks = distBetweenRotatedRect(rotation, tickBound.width, tickBound.height, orient);
                    return tickArea / (distBwTicks.tickSpacing + mathAbs(margin.left) + mathAbs(margin.right));
                }
            }

            if (side === "height") {
                temp = tickBound.height + mathAbs(margin.top) + mathAbs(margin.bottom);
            } else {
                temp = tickBound.width + mathAbs(margin.left) + mathAbs(margin.right);
            }
            return tickArea / temp;
        },

        recalculatePlotSize: function(chartObj, oldPlotSize, tickSize, side, tickCount, margin, padding, outerPadding ){
            var temp, v, p;
            if (side === "height") {
                temp = tickSize[side] + margin.top + margin.bottom;
                v = temp * tickCount;
                oldPlotSize = [v, 0]
            } else {
                temp = tickSize[side] + margin.left + margin.right;
                v = temp * tickCount;
                oldPlotSize = [0,v]
            }
            p = getOuterPaddingInPixel(oldPlotSize, tickCount, padding, outerPadding, temp)
            v += p;
            chartObj.systemConf.chart[side] = v;
        }

    };

    axis.Ticks = Ticks;
    axis.AxisLabel = AxisLabel;

})($ZC);