(function() {

    /**
     *
     * @param {Chart} chartInstance
     * @param {Array} tickValues
     * @param {Array} groupOrder
     * @param {Boolean} isDiscrete
     * @param {Object} columnInfo
     * @constructor
     */

    function HierarchicalTimeDataAdapter(chartInstance, tickValues, groupOrder, isDiscrete, columnInfo) {

        HierarchicalDataAdapter.call(this, chartInstance);

        this.dataType = TIME_DATATYPE;
        this.chart = chartInstance;
        this.columnInfo = columnInfo;

        if (isDiscrete) {
            this.groupByDiscreteTime = true;
        } else {
            this.groupByContinuousTime = true;
        }

        if(isArray(groupOrder)) {
            this.groupOrder = groupOrder.slice(0).reverse();
        } else if (groupOrder === "AUTO") {
            this.groupOrder = HierarchicalTimeDataAdapter.getAutoGroupOrder(columnInfo.time.subfunction);
        }

        this.groupByTime = true; // common

        this.update(tickValues);
    }


    var months = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"],
        days = ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'],
        defaultGroupOrder = ["YEAR", "MONTH", "WEEK", "DAY"],
        formattingFunctions = {
            YEAR: "YEAR",
            ABSQUARTER: "YEAR",
            MONTH: "MONTH",
            ABSMONTH: "MONTH",
            WEEK: "WEEK",
            ABSWEEK: "WEEK",
            ABSDAY: "DAY"
        };

    /**
     *
     * @param {Date} d
     * @returns {string}
     */

    function getMonth(d) {
        return months[d.getMonth()]
    }


    /**
     *
     * @param {Date} d
     * @returns {string}
     */
    function getDay(d) {
        return days[d.getDay()];
    }

    /**
     *
     * @param {Date} d
     * @returns {number}
     */
    function getWeek(d) {
        var firstJanuary = new Date(d.getFullYear(), 0, 1);
        return mathCeil((((d - firstJanuary) / 86400000) + firstJanuary.getDay() + 1) / 7);
    }

    /**
     *
     * @param array
     * @param callback
     * @returns {*}
     */

    function groupBy(array, callback) {
        var key = null;
        return array.reduce(function(r, a) {
            key = callback.call(null, a);
            r[key] = r[key] || [];
            r[key].push(a);
            return r;
        }, {});
    }

    /**
     *
     * @param tickValues
     * @param columnInfo
     */
    function getDateFromTickValues(tickValues, columnInfo) {
        return tickValues.map(function(d) {
            return generateDateVal(d, null, columnInfo);
        });
    }

    /**
     *
     * @param data
     */

    function convertDateToObject(data) {

        var converter = function(point) {
            return {
                YEAR: point.getFullYear(),
                MONTH: getMonth(point),
                WEEK: getWeek(point),
                DAY: getDay(point),
                timestamps: point.getTime(),
                data: point
            };
        };
        return data.map(converter);
    }


    /**
     *
     * @param object
     * @param leaf
     * @param depth
     * @param isDiscreteTime
     * @returns {Array}
     */

    function toHierarchy(object, leaf, depth, isDiscreteTime) {
        var tree = [];
        for (var key in object) {
            if (object.hasOwnProperty(key)) {
                var entry = {
                    depth: depth
                };

                tree.push(entry);

                if (leaf) {
                    entry.leaf = true;
                    entry.value = isDiscreteTime ? object[key].data.getTime(): object[key].data;
                    // entry.value = object[key].data.getTime();
                    entry.point = object[key].data.getTime();

                } else {
                    entry.value = key;
                    entry.point = object[key][0].data;
                    entry.children = object[key]
                }
            }
        }
        return tree
    }

    /**
     *
     * @type {{update: update, process: process, getPossibleGroupBy: getPossibleGroupBy}}
     */

    var timeAxisGroup = {


        update: function(tickValues) {

            var layout = this;

            if(layout.groupByDiscreteTime) {
                tickValues = pick(layout.chart.dataObject.findXCategories(), tickValues);
            }

            layout.dataCount = 0;
            layout.tickValues = getDateFromTickValues(tickValues, layout.columnInfo);
            layout.process();
            layout.setDepth();
        },

        process: function() {

            var layout = this,
                data = layout.tickValues,
                index = 0,
                groupOrder = pick(layout.groupOrder, defaultGroupOrder),
                isDiscreteTime = layout.groupByDiscreteTime;

            // var groupByObject = layout.getPossibleGroupBy(data, groupOrder, index);
            // groupOrder = groupOrder.slice(groupOrder.indexOf(groupByObject));
            // axes count need to be updated
            // layout.chart.axes.x.count = groupOrder.length + 1;
            // layout.chart.ismultix = groupOrder.length > 0;

            var object = convertDateToObject(data, index),
                order = groupOrder.slice(0),
                key = order.shift(),
                yearlyGrouped = groupBy(object, function(date) {
                    return date[key];
                }),
                currentDepth = 0,
                storage = toHierarchy(yearlyGrouped, false, currentDepth, isDiscreteTime);



            function group(currentObj, keys, depth) {

                var key = keys.shift(), grouped;
                currentObj.forEach(function(d) {
                    if (key) {
                        grouped = groupBy(d.children, function(data) {
                            return data[key]
                        });

                        d.children = toHierarchy(grouped, false, depth, isDiscreteTime);
                        group(d.children, keys.slice(0), depth + 1);
                    } else {
                        d.children = toHierarchy(d.children, true, depth, isDiscreteTime);
                    }
                });

                return currentObj;
            }

            layout.categories = group(storage, order.slice(0), currentDepth + 1);
        },

        /**
         *
         * @param {[Date]} data array of dates
         * @param {array} groupOrder
         * @returns {string}
         */

        getPossibleGroupBy: function(data, groupOrder) {
            var startTime = data[0],
                endTime = data[data.length - 1],
                isSameYear = startTime.getFullYear() === endTime.getFullYear(),
                groupByObject = groupOrder[0];

            var timeDiff = data[data.length - 1] - data[0];
            var absoluteTime = millisecondToAbsoluteTime(timeDiff);

            for (var property in absoluteTime) {
                if (absoluteTime.hasOwnProperty(property)) {
                    if (absoluteTime[property] > 1 || (property === "YEAR" && !isSameYear)) {
                        groupByObject = property;
                        break;
                    }
                }
            }
            return groupByObject;
        }

    };


    /**
     *
     * @param {String} formattingFunction
     * @returns {Array.<string>}
     */
    HierarchicalTimeDataAdapter.getAutoGroupOrder = function(formattingFunction) {
        var index = defaultGroupOrder.indexOf(formattingFunctions[formattingFunction]);
        return defaultGroupOrder.slice(0, index);
    };


    // inherit from HierarchicalDataAdapter
    inherit(HierarchicalTimeDataAdapter, HierarchicalDataAdapter);

    // extend timeAxisGroup
    extend(HierarchicalTimeDataAdapter.prototype, timeAxisGroup);

    $ZC.HierarchicalTimeDataAdapter = HierarchicalTimeDataAdapter;

})();
