//$Id$
function attachevents(ele, mouseHandlers, touchHandlers, options, namespace) {
    /*
     * This function attaches event listeners to the element. For touch devices, events will be mapped with touchstart,touchmove and touchend.
     * ele -> To which element the listeners will be attached
     * namespace -> To register multiple listeners for the same event type, the optional namespace is used. (e.g) : "zoom" or null. Refer github.com/mbostock/d3/wiki/Selections#on
     * mouseHandlers -> An object of event type and listener. (e.g) : { "click" : function(){}, "mousemove" : function(){} }
     * touchleave -> whether to handle touch leave or not
     */
     options = options || {};
     namespace = options.namespace || namespace;

    if ($Browser.isMouseDevice && mouseHandlers) {
        attachMouseEvents(ele, namespace, mouseHandlers);
    }
    if ($Browser.isTouchDevice && touchHandlers) {
        attachTouchEvents(ele, namespace, touchHandlers, options);
    }
}

function detachevent(ele, namespace, handlers) {
    namespace = (namespace != null && namespace != "") ? "." + namespace : "";
    if ($Browser.isMouseDevice) {
        if (namespace !== "") {
            ele.on(namespace, null);
        } else {
            for (var attr in handlers) {
                ele.on(attr, null);
            }
        }
        handlers = null;
    }
}

function getEventObject(e) {
    return e.changedTouches ? e.changedTouches[0] : e;
}

function attachMouseEvents(ele, namespace, handlers) {
    if (namespace) {
        namespace = "." + namespace;
        var updatedHandlers = {};
        for (var attr in handlers) {
            updatedHandlers[(attr + namespace)] = handlers[attr];
        }
        handlers = updatedHandlers;
    }

    ele.on(handlers);
}

function attachTouchEvents(ele, namespace, handlers, options) {
    var touchleave = options.touchleave,
        target = options.target,
        chartObj = options.chartObj,
        plotEventOptions = (chartObj && chartObj.systemConf.chart.plot.events) || {},
        doPreventDefault = options.preventDefault !== false;
    var isFocusedTouchmove = !((target === "plot") && plotEventOptions.isFocusedTouchmove == false);
    namespace = (namespace != null) ? "." + namespace : "";

    var MINDISTANCE = 30,
        SINGLETAPDURATION = 150,
        DOUBLETAPDURATION = 500,
        FOCUSEDMOVEDURATION = 100,
        TOUCHMOVETIMER = 50,
        TAPHOLDDURATION = 500;
    var prevElement, touchstart = {},
        lastMove = 0,
        lastTap = 0,
        tapDifference = DOUBLETAPDURATION - SINGLETAPDURATION,
        firstMove = null,
        isDoubletouchStarted = false,
        tapholdDetected,
        tapHoldTimer;

    var mousemoveFn = handlers.touchmove,
        touchendFn = handlers.touchend,
        tapFn = handlers.tap,
        dbltapFn = handlers.dbltap,
        singletouchmoveFn = handlers.singletouchmove,
        doubletouchmoveFn = handlers.doubletouchmove,
        tapholdFn,
        tapholdendFn;

        if (target === "plot" && getEventConf(null, chartObj.systemConf.chart.plot.plotoptions[chartObj.chartTypes[0].name], chartObj, "taphold")) {
            tapholdFn = handlers.tap;
            tapholdendFn = touchendFn;
        } else {
            tapholdFn = handlers.taphold;
            tapholdendFn = handlers.tapholdend;
        }

    var touchstartProp = "touchstart" + namespace,
        touchmoveProp = "touchmove" + namespace,
        touchendProp = "touchend" + namespace;

    var updatedHandlers = {};
    updatedHandlers[touchstartProp] = function() {
        if (plotEventOptions.enabled == false) {
            return;
        }

        var touchObject = d3.event.touches;

        if (touchObject.length > 1) {
            return;
        }

        touchstart.time = new Date();
        touchstart.location = getMousePosition(this, d3.event);
        firstMove = null;
        isDoubletouchStarted = false;
        tapholdDetected = false;

        if (tapholdFn) {
            var e = d3.event;
            var arg = arguments;
            var self = this;
            tapHoldTimer = setTimeout(function(){
                d3.event = e;
                if (chartObj) {
                    chartObj.zcEventType = "taphold"; //change
                }

                tapholdFn.apply(self, arg);

                tapholdDetected = true;
            }, TAPHOLDDURATION);
        }
    }
    updatedHandlers[touchmoveProp] = function() {
        if (plotEventOptions.enabled == false) {
            return;
        }

        var touchObject = d3.event.touches;

        if (tapholdFn) {
            var currentLocation = getMousePosition(this, d3.event);
            var distance = (touchstart.location != null) ? getEuclideanDistance(touchstart.location[0], touchstart.location[1], currentLocation[0], currentLocation[1]) : null;

            if (distance > MINDISTANCE) {
                clearTimeout(tapHoldTimer);
            } else {
                if (tapholdDetected) {
                    preventDefault();
                    return false;
                }
            }
        }

        if (touchObject.length > 1) {
            if ((touchObject.length == 2) && doubletouchmoveFn) {
                isDoubletouchStarted = true;
                doubletouchmoveFn.apply(this, arguments);
                preventDefault();
            }
            return;
        }

        var currentTime = new Date().getTime();
        if (!firstMove) {
            firstMove = currentTime;
        }

        // Touchleave event is not available. So finding the element to which the touchmove would be called is necessary. Call the mouseout of previous element and mousemove of the current element
        if (touchleave && document.elementFromPoint != null) {
            var timer = (lastMove == 0 || (currentTime - lastMove) > TOUCHMOVETIMER);
            var curTouch = touchObject[0];
            var currentEle = document.elementFromPoint((curTouch.pageX - window.pageXOffset), (curTouch.pageY - window.pageYOffset));
            if (!timer || (currentEle.localName === "svg") || (currentEle.id.match(/tooltip/gi) != null)) {
                preventDefault();
                return;
            }

            if (currentEle != this) {
                // Call mouse out of the previous element
                if (prevElement != currentEle) {
                    var touchEndElement = prevElement || this;
                    var element = d3.select(touchEndElement);
                    var endEvent = element.on(touchendProp),
                        d = element.datum(),
                        identity = element.attr("id") || element.attr("class");
                    var i = (identity != null) ? identity.match(/(\d)+$/gi) : null;
                    i = (i != null && i[0] != null) ? parseInt(i[0]) : null;
                    if (endEvent != null && i != null) {
                        endEvent.call(touchEndElement, d, i);
                    }
                }
                // Call mouse move of the current element
                var element = d3.select(currentEle);
                var moveEvent = element.on(touchmoveProp),
                    d = element.datum(),
                    identity = element.attr("id") || element.attr("class");
                var i = (identity != null) ? identity.match(/(\d)+$/gi) : null;
                i = (i != null && i[0] != null) ? parseInt(i[0]) : null;
                if (moveEvent != null && i != null) {
                    moveEvent.call(currentEle, d, i);
                    prevElement = currentEle;
                    lastMove = currentTime;
                }
                return;
            }
        }

        if (!(isFocusedTouchmove && ((firstMove - touchstart.time) < FOCUSEDMOVEDURATION))) {
            if (singletouchmoveFn) {
                if (chartObj) {
                    chartObj.zcEventType = "touchmove"; //change
                }

                singletouchmoveFn.apply(this, arguments);
                preventDefault();
            }
        }

    };
    updatedHandlers[touchendProp] = function() {
        if (plotEventOptions.enabled == false) {
            return;
        }

        var touchObject = d3.event.touches;

        if (tapholdFn) {
            clearTimeout(tapHoldTimer);
            if (tapholdDetected) {
                tapholdDetected = false;
                preventDefault();
                if (tapholdendFn) {
                    tapholdendFn.apply(this, arguments);
                }
                return false;
            }
        }

        if (touchObject.length > 1) {
            return;
        }

        if (isDoubletouchStarted && doubletouchmoveFn && touchendFn) {
            isDoubletouchStarted = false;
            touchendFn.apply(this, arguments);
            preventDefault();
            return;
        }

        var currentTime = new Date(),
            currentLocation = getMousePosition(this, d3.event);
        var duration = currentTime - touchstart.time;
        var distance = (touchstart.location != null) ? getEuclideanDistance(touchstart.location[0], touchstart.location[1], currentLocation[0], currentLocation[1]) : null;
        var tapLength = currentTime - lastTap;

        if (tapDifference > 0 && tapLength <= tapDifference && tapLength > 0 && distance <= MINDISTANCE) { // Double tap event
            if (dbltapFn) {
                if (chartObj){
                    chartObj.zcEventType = "doubletap"; //change
                }

                dbltapFn.apply(this, arguments);
                preventDefault();
            }
            lastTap = 0;
        } else if (duration <= SINGLETAPDURATION && distance <= MINDISTANCE) { // Single tap event
            if (tapFn) {
                if (chartObj) {
                    chartObj.zcEventType = "tap"; //change
                }

                tapFn.apply(this, arguments);
                preventDefault();
            }
        } else {
            if (touchendFn) {
                touchendFn.apply(this, arguments);
            }
        }
        lastTap = currentTime;
    };

    ele.on(updatedHandlers)

    function getEuclideanDistance(x1, y1, x2, y2) {
        return Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2));
    }

    function preventDefault() {
        if (doPreventDefault) {
            d3.event.preventDefault();
        }
    }
}